# -*- coding: utf-8 -*-

# QState

class QState_Error_Initialize(Exception):
    def __str__(self):
        return "QState: fail to initialize"

class QState_Error_GetCmp(Exception):
    def __str__(self):
        return "QState: fail to get cmp"

class QState_Error_Show(Exception):
    def __str__(self):
        return "QState: fail to show"

class QState_Error_Clone(Exception):
    def __str__(self):
        return "QState: fail to clone"

class QState_Error_InnerProduct(Exception):
    def __str__(self):
        return "QState: fail to inner product"

class QState_Error_TensorProduct(Exception):
    def __str__(self):
        return "QState: fail to tensor product"

class QState_OutOfBound(Exception):
    def __str__(self):
        return "QState: out of bound"

class QState_TooManyArguments(Exception):
    def __str__(self):
        return "QState: too many arguments"

class QState_NeedMoreArguments(Exception):
    def __str__(self):
        return "QState: need more arguments"

class QState_SameQubitID(Exception):
    def __str__(self):
        return "QState: same qubit id"
    
class QState_UnknownQgateKind(Exception):
    def __str__(self):
        return "QState: unknown qgate kind"

class QState_Error_OperateQgate(Exception):
    def __str__(self):
        return "QState: fail to operate qgate"

class QState_Error_Evolve(Exception):
    def __str__(self):
        return "QState: fail to evolve"

class QState_Error_Expect(Exception):
    def __str__(self):
        return "QState: fail to expect"

class QState_Error_Apply(Exception):
    def __str__(self):
        return "QState: fail to apply"

class QState_Error_Bloch(Exception):
    def __str__(self):
        return "QState: fail to bloch"

class QState_Error_Reset(Exception):
    def __str__(self):
        return "QState: fail to reset"

class QState_Error_FreeAll(Exception):
    def __str__(self):
        return "QState: fail to free all"

class QState_Error_AddMethods(Exception):
    def __str__(self):
        return "QState: fail to add methods"

class QState_Error_OperateQcirc(Exception):
    def __str__(self):
        return "QState: fail to operate qcirc"

# MData

class MData_Error_Initialize(Exception):
    def __str__(self):
        return "MData: fail to initialize"

class MData_Error_GetFrq(Exception):
    def __str__(self):
        return "MData: fail to get frq"

class MData_Error_Show(Exception):
    def __str__(self):
        return "MData: fail to show"

class MData_Error_GetMeasuredData(Exception):
    def __str__(self):
        return "MData: fail to get measured data (probably not measured yet)"

# Observable

class Observable_Error_Initialize(Exception):
    def __str__(self):
        return "Observable: fail to initialize"

# DensOp

class DensOp_Error_Initialize(Exception):
    def __str__(self):
        return "DensOp: fail to initialize"

class DensOp_Error_Show(Exception):
    def __str__(self):
        return "DensOp: fail to show"

class DensOp_Error_Clone(Exception):
    def __str__(self):
        return "DensOp: fail to clone"

class DensOp_Error_Add(Exception):
    def __str__(self):
        return "DensOp: fail to add"

class DensOp_Error_Mul(Exception):
    def __str__(self):
        return "DensOp: fail to mul"

class DensOp_Error_GetElm(Exception):
    def __str__(self):
        return "DensOp: fail to get_elm"

class DensOp_Error_Trace(Exception):
    def __str__(self):
        return "DensOp: fail to trace"

class DensOp_Error_SqTrace(Exception):
    def __str__(self):
        return "DensOp: fail to square trace"

class DensOp_Error_PaTrace(Exception):
    def __str__(self):
        return "DensOp: fail to partial trace"

class DensOp_Error_Apply(Exception):
    def __str__(self):
        return "DensOp: fail to apply"

class DensOp_Error_Probability(Exception):
    def __str__(self):
        return "DensOp: fail to probability"

class DensOp_Error_Instrument(Exception):
    def __str__(self):
        return "DensOp: fail to instrument"

class DensOp_Error_Fidelity(Exception):
    def __str__(self):
        return "DensOp: fail to get fidelity"

class DensOp_Error_Distance(Exception):
    def __str__(self):
        return "DensOp: fail to get distance"

class DensOp_Error_Spectrum(Exception):
    def __str__(self):
        return "DensOp: fail to get spectrum"

class DensOp_Error_Entropy(Exception):
    def __str__(self):
        return "DensOp: fail to get entropy"

class DensOp_Error_OperateQGate(Exception):
    def __str__(self):
        return "DensOp: fail to operate qgate"

class DensOp_Error_TensorProduct(Exception):
    def __str__(self):
        return "DensOp: fail to tensor product"
    
class DensOp_Error_Reset(Exception):
    def __str__(self):
        return "DensOp: fail to reset"
    
class DensOp_Error_FreeAll(Exception):
    def __str__(self):
        return "DensOp: fail to free all"

class DensOp_Error_AddMethods(Exception):
    def __str__(self):
        return "DensOp: fail to add methods"

# Stabilizer

class Stabilizer_Error_Initialize(Exception):
    def __str__(self):
        return "Stabilizer: fail to initialize"

class Stabilizer_Error_Clone(Exception):
    def __str__(self):
        return "Stabilizer: fail to clone"

class Stabilizer_Error_AddMethos(Exception):
    def __str__(self):
        return "Stabilizer: fail to add methods"

class Stabilizer_Error_SetPauliFac(Exception):
    def __str__(self):
        return "Stabilizer: fail to set pauli factor."

class Stabilizer_Error_GetPauliFac(Exception):
    def __str__(self):
        return "Stabilizer: fail to get pauli factor."

class Stabilizer_Error_SetPauliOp(Exception):
    def __str__(self):
        return "Stabilizer: fail to set pauli op."

class Stabilizer_Error_GetPauliOp(Exception):
    def __str__(self):
        return "Stabilizer: fail to get pauli op."

class Stabilizer_Error_GetRank(Exception):
    def __str__(self):
        return "Stabilizer: fail to get rank."

class Stabilizer_Error_OperateQgate(Exception):
    def __str__(self):
        return "Stabilizer: fail to operate qgate."

class Stabilizer_Error_GetRank(Exception):
    def __str__(self):
        return "Stabilizer: fail to get rank."

class Stabilizer_Error_Measure(Exception):
    def __str__(self):
        return "Stabilizer: fail to measure qubits."

class Stabilizer_Error_FreeAll(Exception):
    def __str__(self):
        return "Stabilizer: fail to free all"

class Stabilizer_Error_OperateQcirc(Exception):
    def __str__(self):
        return "Stabilizer: fail to operate qcirc"

# Backend

class Backend_Error_NameNotSupported(Exception):
    def __str__(self):
        return "Backend: name is not supported"

class Backend_Error_DeviceNotSupported(Exception):
    def __str__(self):
        return "Backend: device is not supported"

class Backend_Error_NameNotSpecified(Exception):
    def __str__(self):
        return "Backend: name is not specified"

# QCirc

class QCirc_Error_Initialize(Exception):
    def __str__(self):
        return "QCirc: fail to initialize"

class QCirc_Error_Copy(Exception):
    def __str__(self):
        return "QCirc: fail to copy"

class QCirc_Error_Merge(Exception):
    def __str__(self):
        return "QCirc: fail to merge"

class QCirc_Error_AppendGate(Exception):
    def __str__(self):
        return "QCirc: fail to append gate"

class QCirc_Error_KindFirst(Exception):
    def __str__(self):
        return "QCirc: fail to kind first"

class QCirc_Error_PopGate(Exception):
    def __str__(self):
        return "QCirc: fail to pop gate"

# CMem

class CMem_Error_Initialize(Exception):
    def __str__(self):
        return "CMem: fail to initialize"

class CMem_Error_Clone(Exception):
    def __str__(self):
        return "CMem: fail to clone"




import os
import subprocess
import sys

def main():
    args = sys.argv
    os.environ['LD_LIBRARY_PATH'] = os.path.dirname(__file__) + '/lib'
    os.environ['PATH'] = '~/.local/bin'
    raise SystemExit(subprocess.call([os.path.dirname(os.path.abspath(__file__)) + '/qlazy'] + args[1:]))
    
if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import ctypes
from collections import Counter
from ctypes.util import find_library
from fractions import Fraction
import warnings
import ctypes
import re
import random
import math
import pickle

from qlazy.config import *
from qlazy.error import *

def string_to_args(s):  # for from_qasm

    token = s.split(' ')
    if len(token) == 1:
        args = token
    elif len(token) > 1:
        args = [token[0], ' '.join(token[1:])]
    else:
        raise ValueError("can't split string {}.".format(s))

    return args
            
class QCirc(ctypes.Structure):
    """ Quantum Circuit

    Attributes
    ----------
    qubit_num : int
        qubit number of the quantum state (= log(state_num)).
    cmem_num : int
        number of the classical register.
    gate_num : int
        number of gates in the quantum circuit.
    first: object
        first gate of the quantum circuit.
    last: object
        last gate of the quantum circuit.

    """
    _fields_ = [
        ('qubit_num', ctypes.c_int),
        ('cmem_num', ctypes.c_int),
        ('gate_num', ctypes.c_int),
        ('first', ctypes.c_void_p),
        ('last', ctypes.c_void_p),
    ]

    def __new__(cls, **kwargs):
        """
        Parameters
        ----------
        None

        Returns
        -------
        qcirc : instance (QCirc)

        """
        obj = qcirc_init()
        qcirc = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
        return qcirc

    def __str__(self):

        return self.to_string()

    def __add__(self, qc):
        """
        Parameters
        ----------
        qc : instance of QCirc
            quantum circuit (merged)

        Returns
        -------
        qcirc : instance of QCirc
            quantum circuit (result)

        """
        qcirc = self.merge(qc)
        return qcirc

    def __iadd__(self, qc):
        """
        Parameters
        ----------
        qc : instance of QCirc
            quantum circuit (merged)

        Returns
        -------
        qcirc : instance of QCirc
            quantum circuit (result)

        """
        qcirc = self.merge(qc)
        return qcirc

    def __eq__(self, qc):
        """
        Parameters
        ----------
        qc : instance of QCirc
            quantum circuit (merged)

        Returns
        -------
        ans : bool
            equal or not

        """
        ans = self.is_equal(qc)
        return ans

    def __ne__(self, qc):
        """
        Parameters
        ----------
        qc : instance of QCirc
            quantum circuit (merged)

        Returns
        -------
        ans : bool
            equal or not

        """
        ans = not self.is_equal(qc)
        return ans

    def to_string(self):
        """
        get string of the circuit (qlazy format).

        Parameters
        ----------
        None

        Returns
        -------
        qcirc_str : str

        """
        qc = self.clone()
        qcirc_str = ""
        while True:
            kind = qc.kind_first()
            if kind is None:
                break
            else:
                (kind, qid, para, c, ctrl) = qc.pop_gate()
                term_num = get_qgate_qubit_num(kind)
                if (kind == MEASURE) or (kind == RESET):
                    term_num = 1
                para_num = get_qgate_param_num(kind)

                gate_str = GATE_STRING[kind]
                qid_str = " ".join(map(str, [qid[i] for i in range(term_num)]))
                qid_str.strip()

                if para_num == 0:
                    para_str = ""
                else:
                    para_str = ",".join(map(str, [para[i] for i in range(para_num)]))
                    para_str = "(" + para_str+ ")"
                
                if c is None:
                    c_str = ""
                else:
                    c_str = "-> {}".format(c)

                if ctrl is None:
                    ctrl_str = ""
                else:
                    ctrl_str = ", ctrl = {}".format(ctrl)
                    
                qcirc_str += "{0:}{2:} {1:} {3:}{4:}\n".format(gate_str, qid_str, para_str, c_str, ctrl_str)
                
        return qcirc_str.strip()
    
    @classmethod
    def from_qasm(cls, string):
        """
        get QCirc instance from OpenQASM 2.0 string.

        Parameters
        ----------
        None

        Returns
        -------
        qcirc : instance of QCirc

        Notes
        -----
        Non-unitary gates (measure, reset) and user customized gates are not supported. Supported gates are 
        'x', 'y', 'z', 'h', 's', 'sdg', 't', 'tdg', 'cx', 'cz', 'ch', 'rx', 'rz', 'crz'.

        """
        line_list = string.split('\n')

        # line #0
        line_count = 0
        args = string_to_args(line_list[line_count])
        if args[0] == 'OPENQASM' and args[1] == '2.0;':
            line_count += 1
        else:
            raise ValueError("line #0 must be 'OPENQASM 2.0;'")

        # line #1
        args = string_to_args(line_list[line_count])
        if args[0] == 'include' and args[1] == '"qelib1.inc";':
            line_count += 1
        else:
            raise ValueError("""line #1 must be 'include "qelib1.inc;"' """)

        # line #2 (#3)
        args = string_to_args(line_list[line_count])
        if args[0] == 'qreg' and re.match('q', args[1]).group() == 'q':
            qubit_num = int(re.sub("q\[|\];", "", args[1]))
            line_count += 1
        else:
            raise ValueError("""line #2 must be 'qreg q[<int>];"' """)

        # line (#3) #4 ...
        qcirc = cls()
        for i in range(line_count, len(line_list)):
            args = string_to_args(line_list[i])
            if args[0] == '': continue

            if args[0] in ('measure', 'reset'):
                raise ValueError("sorry, 'measure', 'reset' is not supported.")
            
            if args[0] in ('x', 'y', 'z', 'h', 's', 'sdg', 't', 'tdg'):
                res = re.search("q\[|\];", args[1])
                if res is not None:
                    q = int(re.sub("q\[|\];", "", args[1]))
                else:
                    raise ValueError("argument '{}' is not valid for '{}' gate.".format(args[1], args[0]))
                
                if args[0] == 'x':
                    qcirc.x(q)
                elif args[0] == 'z':
                    qcirc.z(q)
                elif args[0] == 'h':
                    qcirc.h(q)
                elif args[0] == 's':
                    qcirc.s(q)
                elif args[0] == 'sdg':
                    qcirc.s_dg(q)
                elif args[0] == 't':
                    qcirc.t(q)
                elif args[0] == 'tdg':
                    qcirc.t_dg(q)
                    
            elif args[0] in ('cx', 'cz', 'ch'):
                res = re.search("q\[[0-9]+\],\s*q\[[0-9]+\];", args[1])
                if res is not None:
                    qubits = args[1].split(',')
                    q0 = int(re.sub("q\[|\]", "", qubits[0]))
                    q1 = int(re.sub("q\[|\];", "", qubits[1]))
                else:
                    raise ValueError("argument '{}' is not valid for '{}' gate.".format(args[1], args[0]))
                
                if args[0] == 'cx':
                    qcirc.cx(q0, q1)
                elif args[0] == 'cz':
                    qcirc.cz(q0, q1)
                elif args[0] == 'ch':
                    qcirc.ch(q0, q1)
                    
            elif (re.match('rx', args[0]) is not None or
                  re.match('rz', args[0]) is not None):

                res = re.search("q\[|\];", args[1])
                if res is not None:
                    q = int(re.sub("q\[|\];", "", args[1]))
                else:
                    raise ValueError("argument '{}' is not valid for '{}' gate.".format(args[1], args[0]))
                
                para_str = [s.strip('*').strip('/') for s in re.sub(".+\(|\)", "", args[0]).split("pi")]
                if para_str[0] == '0':
                    para = 0.
                else:
                    denominator = 1.
                    if para_str[0] == '':
                        numerator = 1.
                    else:
                        numerator = float(para_str[0])
                    if para_str[1] == '':
                        denominator = 1.
                    else:
                        denominator = float(para_str[1])
                    para = numerator / denominator

                if re.match('rx', args[0]) is not None:
                    qcirc.rx(q, phase=para)
                elif re.match('rz', args[0]) is not None:
                    qcirc.rz(q, phase=para)
                
            elif re.match('crz', args[0]) is not None:

                res = re.search("q\[[0-9]+\],\s*q\[[0-9]+\];", args[1])
                if res is not None:
                    qubits = args[1].split(',')
                    q0 = int(re.sub("q\[|\]", "", qubits[0]))
                    q1 = int(re.sub("q\[|\];", "", qubits[1]))
                else:
                    raise ValueError("argument '{}' is not valid for '{}' gate.".format(args[1], args[0]))
                
                para_str = [s.strip('*').strip('/') for s in re.sub(".+\(|\)", "", args[0]).split("pi")]
                if para_str[0] == '0':
                    para = 0.
                else:
                    denominator = 1.
                    if para_str[0] == '':
                        numerator = 1.
                    else:
                        numerator = float(para_str[0])
                    if para_str[1] == '':
                        denominator = 1.
                    else:
                        denominator = float(para_str[1])
                    para = numerator / denominator
                
                if re.match('crz', args[0]) is not None:
                    qcirc.crz(q0, q1, phase=para)
                    
            else:
                raise ValueError("{} gate is not supported".format(args[0]))

        return qcirc
        
    @classmethod
    def from_qasm_file(cls, file_path):
        """
        get QCirc instance from OpenQASM 2.0 file.

        Parameters
        ----------
        file_path: str
            file path name of OpenQASM 2.0 file

        Returns
        -------
        qcirc : instance of QCirc

        Notes
        -----
        Non-unitary gates (measure, reset) and user customized gates are not supported. Supported gates are 
        'x', 'y', 'z', 'h', 's', 'sdg', 't', 'tdg', 'cx', 'cz', 'ch', 'rx', 'rz', 'crz'.

        """
        with open(file_path, mode='r') as f:
            s = f.read()
            
        qcirc = __class__.from_qasm(s)
        return qcirc
    
    def to_qasm(self):
        """
        get OpenQASM 2.0 string of the circuit.

        Parameters
        ----------
        None

        Returns
        -------
        qcirc_str : str

        """
        qc = self.clone()

        # header and include file
        qcirc_str = """OPENQASM 2.0;\n"""
        qcirc_str += """include "qelib1.inc";\n"""

        # definition of qreg, creg
        qcirc_str += """qreg q[{}];\n""".format(self.qubit_num)
        for i in range(self.cmem_num):
            qcirc_str += """creg c{}[1];\n""".format(i)
        
        # description of each gate operation
        while True:
            kind = qc.kind_first()
            if kind is None:
                break
            else:
                (kind, qid, para, c, ctrl) = qc.pop_gate()
                term_num = get_qgate_qubit_num(kind)
                if (kind == MEASURE) or (kind == RESET):
                    term_num = 1

                para_num = get_qgate_param_num(kind)
                para_frac = [Fraction(str(p)) for p in para]

                gate_str = GATE_STRING_QASM[kind]
                qid_str = ",".join(["q[" + str(qid[i]) + "]" for i in range(term_num)])
                qid_str.strip()

                if para_num == 0:
                    if kind == CONTROLLED_S:
                        para_str = "(pi/2)"
                    elif kind == CONTROLLED_S_:
                        para_str = "(-pi/2)"
                    elif kind == CONTROLLED_T:
                        para_str = "(pi/4)"
                    elif kind == CONTROLLED_T_:
                        para_str = "(-pi/4)"
                    else:
                        para_str = ""
                else:
                    para_str_list = []
                    for i, p in enumerate(para_frac):
                        if i >= para_num:
                            break
                        if p.numerator == 0:
                            para_str_list.append("0")
                        elif p.numerator == 1:
                            para_str_list.append("pi/"+ str(p.denominator))
                        else:
                            para_str_list.append(str(p.numerator) + "*pi/" + str(p.denominator))
                            
                    para_str = ",".join(para_str_list)
                    para_str = "(" + para_str+ ")"
                
                if c is None:
                    c_str = ""
                else:
                    c_str = " -> c{}[0]".format(c)

                if ctrl is None:
                    ctrl_str = ""
                else:
                    ctrl_str = "if(c{}==1) ".format(ctrl)
                    
                qcirc_str += "{4:}{0:}{2:} {1:}{3:};\n".format(gate_str, qid_str, para_str, c_str, ctrl_str)
                
        return qcirc_str.strip()

    def to_qasm_file(self, file_path):
        """
        write to OpenQASM 2.0 file.

        Parameters
        ----------
        file_path: str
            file path name of OpenQASM 2.0 file

        Returns
        -------
        None

        """
        s = self.to_qasm()
        with open(file_path, mode='w') as f:
            f.write(s)
        
    def clone(self):
        """
        clone quantum circuit.

        Parameters
        ----------
        None

        Returns
        -------
        qcirc : instance of QCirc
            quantum circuit

        """
        obj = qcirc_copy(self)
        qcirc = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return qcirc

    def merge(self, qc):
        """
        merge quantum circuit.

        Parameters
        ----------
        qc : instance of QCirc
            quantum circuit (merged)

        Returns
        -------
        qcirc : instance of QCirc
            quantum circuit (result)

        """
        obj = qcirc_merge(self, qc)
        qcirc = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return qcirc

    def is_equal(self, qc):
        """
        eaual or not.

        Parameters
        ----------
        qc : instance of QCirc
            quantum circuit (merged)

        Returns
        -------
        qcirc : instance of QCirc
            quantum circuit (result)

        """
        ans = qcirc_is_equal(self, qc)
        return ans

    def kind_first(self):
        """
        get kind of first gate of the circuit.

        Parameters
        ----------
        None

        Returns
        -------
        kind_list : int
            kind of first quantum gate of quantum circuit

        Note
        ----
        return None if none of gates included

        """
        kind = qcirc_kind_first(self)
        return kind

    def kind_list(self):
        """
        get list of kinds from the circuit.

        Parameters
        ----------
        None

        Returns
        -------
        kinds : list of kind
            kinds of the quantum circuit

        Note
        ----
        return None if none of gates included

        """
        qc = self.clone()
        kind_list = []
        while True:
            kind = qc.kind_first()
            if kind is None: break
            kind_list.append(kind)
            qc.pop_gate()
        
        return kind_list
    
    def get_gates(self):
        """
        get list of gates from the circuit.

        Parameters
        ----------
        None

        Returns
        -------
        gates : list of dict
            gates of the quantum circuit
            gates = [{'kind':kind, 'qid':qid, 'phase':phase, 'cid':cid, 'ctrl':ctrl}, ...]
            - kind: gate name
            - qid: qubit id
            - phase: phase parameter (non-zero only for rotation gates)
            - cid: classical register id (set only for measurement gate)
            - ctrl: classical register id to control gate operation

        """
        qc = self.clone()
        gates = []
        while True:
            kind = qc.kind_first()
            if kind is None:
                break
            else:
                (kind, qid, para, c, ctrl) = qc.pop_gate()
                qid = [q for q in qid if q >= 0]
                if c is None: cid = None
                else: cid = [c]
                gates.append({'kind': GATE_STRING[kind], 'qid': qid, 'phase': para[0], 'cid': cid, 'ctrl': ctrl})

        return gates

    def add_gates(self, gates=None):
        """
        add list of gates to the circuit.

        Parameters
        ----------
        gates : list of dict
            gates of the quantum circuit
            gates = [{'kind':kind, 'qid':qid, 'phase':phase, 'cid':cid, 'ctrl':ctrl}, ...]
            - kind: gate name
            - qid: qubit id
            - phase: phase parameter (non-zero only for rotation gates)
            - cid: classical register id (set only for measurement gate)
            - ctrl: classical register id for controlling gate operation

        Returns
        -------
        self: instance of QCirc
            circuit after adding gates

        """
        if gates is None:
            raise ValueError("gates must be specified.")
        for g in gates:
            kind = GATE_KIND[g['kind']]
            qid = g['qid']
            para = [g['phase'], 0.0, 0.0]
            if g['cid'] is None:
                c = None
            else:
                c = g['cid'][0]
            ctrl = g['ctrl']
            self.append_gate(kind, qid, para, c, ctrl)
        return self

    def dump(self, file_path=None):
        """
        dump the circuit

        Parameters
        ----------
        file_path: str
            file path of dump file

        Returns
        -------
        None

        """
        gates = self.get_gates()
        with open(file_path, mode='wb') as f:
            pickle.dump(gates, f)

    @classmethod
    def load(cls, file_path=None):
        """
        load the circuit

        Parameters
        ----------
        file_path: str
            file path of dump file

        Returns
        -------
        qcirc: instance of QCirc
            loaded circuit

        """
        with open(file_path, mode='rb') as f:
            gates = pickle.load(f)
        qcirc = cls().add_gates(gates)
        return qcirc
    
    def get_stats(self):
        """
        get statistics of the circuit.

        Parameters
        ----------
        None

        Returns
        -------
        stats : dict
            {'qubit_num':qubit_num, 'cmem_num':cmem_num, 'gate_num':gate_num, 'gate_freq':gate_freq}
            - qubit_num: number of qubits
            - cmem_num: number of classical bits
            - gate_num: number of gates
            - gate_freq: frequency of gates (Counter)

        """
        gate_list = [GATE_STRING[kind] for kind in self.kind_list()]
        gate_freq = Counter(gate_list)

        gatetype_list = []
        for kind in self.kind_list():
            if is_clifford_gate(kind) == True or is_non_clifford_gate(kind) == True:
                gatetype_list.append('unitary')
                if is_clifford_gate(kind) == True:
                    gatetype_list.append('clifford')
                elif is_non_clifford_gate(kind) == True:
                    gatetype_list.append('non-clifford')
            elif is_measurement_gate(kind) == True or is_reset_gate(kind) == True:
                gatetype_list.append('non-unitary')
            else:
                raise ValueError("unknown gate kind:{}".format(kind))
        gatetype_freq = Counter(gatetype_list)

        stats = {'qubit_num': self.qubit_num, 'cmem_num':self.cmem_num, 'gate_num': len(gate_list),
                 'gate_freq': gate_freq, 'gatetype_freq': gatetype_freq}
        return stats

    @classmethod
    def generate_random_gates(cls, qubit_num=0, gate_num=0, phase=None, prob=None, **kwargs):
        """
        generate circuit including random gates.

        Parameters
        ----------
        qubit_num: int
            number of qubits
        gate_num: int
            numbner of gates
        phase: tupple of float
            phases selected randomly
        prob: dict
            {'x':prob_x, 'z':prob_z, 'h':prob_h, 's':prob_s, 's_dg':prob_s_dg,
             't':prob_t, 't_dg':prob_t_dg, 'rx':prob_rx, 'rz':prob_rz, 'cx':prob_cx,
             'cz':prob_cz, 'ch':prob_ch, 'crz':prob_crz}
            - prob_x: probability of x
            - prob_z: probability of z
            - prob_h: probability of h
            - prob_s: probability of s
            - prob_s_dg: probability of s_dg
            - prob_t: probability of t
            - prob_t_dg: probability of t_dg
            - prob_rx: probability of rx
            - prob_rz: probability of rz
            - prob_cx: probability of cx
            - prob_cz: probability of cz
            - prob_ch: probability of ch
            - prob_crz: probability of crz

        Returns
        -------
        qcirc: instance of QCirc
            generated circuit

        Examples
        --------
        >>> qc = QCirc.generate_random_gates(qubit_num=5, gate_num=100, phase_unit=0.25, prob={'h':3, 'cx':7, 'rz':1})

        Notes
        -----
        * each probability values are normalized so that the sum of the probabilities is 1.0. 
        * Phase parameters of rotation gates are selected randomly in the element of 'phase'.

        """
        if type(qubit_num) != int or type(gate_num) != int or qubit_num < 1 or gate_num < 1:
            raise ValueError("qubit_num and/or gate_num must be positive integer.")
        if phase is not None:
            if type(phase) != float and type(phase) != int and type(phase) != tuple:
                raise ValueError("phase value(s) must be int/float of tuple.")
        
        total_prob = 0.0
        for p in prob.values():
            total_prob += p
        glist = []
        plist = []
        p = 0.0
        for k, v in prob.items():
            if v < 0.0:
                raise ValueError("probability must be positive value.")
            if k  in ('x', 'z', 'h', 's', 's_dg', 't', 't_dg', 'rx', 'rz', 'cx', 'cz', 'ch', 'crz'):
                glist.append(k)
                p += v / total_prob
                plist.append(p)
            else:
                raise ValueError("gate '{}' is not supported.".format(k))

        if plist[-1] != 1.0: plist[-1] = 1.0

        qcirc = cls()

        TRY_MAX = 10  # for random generation
        for n in range(gate_num):
            r = random.random()
            kind = None
            for i, p in enumerate(plist):
                if r <= p:
                    kind = GATE_KIND[glist[i]]
                    break
            term_num = get_qgate_qubit_num(kind)
            para_num = get_qgate_param_num(kind)
            
            if term_num == 1 and para_num == 0:    # 1-qubit gate
                q0 = random.randint(0, qubit_num - 1)
                qcirc.append_gate(kind, [q0])
            elif term_num == 1 and para_num == 1:  # 1-qubit and 1-parameter gate
                q0 = random.randint(0, qubit_num - 1)
                if phase is None: p = 0.0
                elif type(phase) == float or type(phase) == int: p = phase
                else: p = random.choice(phase)
                qcirc.append_gate(kind, [q0], para=[p, 0., 0.])
            elif term_num == 2 and para_num == 0:  # 2-qubit gate
                q0 = random.randint(0, qubit_num - 1)
                q1 = random.randint(0, qubit_num - 1)
                cnt = 0
                while q0 == q1 and cnt < TRY_MAX:
                    q1 = random.randint(0, qubit_num - 1)
                    cnt += 1
                if cnt >= TRY_MAX:
                    raise ValueError("can't generate qubit id for '{}' gate.".format(GATE_STRING[kind]))
                qcirc.append_gate(kind, [q0, q1])
            elif term_num == 2 and para_num == 1:  # 2-qubit and 1-parameter gate
                q0 = random.randint(0, qubit_num - 1)
                q1 = random.randint(0, qubit_num - 1)
                cnt = 0
                while q0 == q1 and cnt < TRY_MAX:
                    q1 = random.randint(0, qubit_num - 1)
                    cnt += 1
                if cnt >= TRY_MAX:
                    raise ValueError("can't generate qubit id for '{}' gate.".format(GATE_STRING[kind]))
                if phase is None: p = 0.0
                elif type(phase) == float or type(phase) == int: p = phase
                else: p = random.choice(phase)
                qcirc.append_gate(kind, [q0, q1], para=[p, 0., 0.])
            else:
                raise ValueError("gate of term_num={}, param_num={} is not supported".format(term_num, para_num))

        return qcirc

    def to_pyzx(self):
        """
        get pyzx's Circuit instance.

        Parameters
        ----------
        None

        Returns
        -------
        zxqc: instance of pyzx's Circuit
            quantum circuit

        Notes
        -----
        Non-unitary gates: measure, reset are not supported.

        """
        import pyzx as zx
        from pyzx import Circuit

        zxqc = Circuit.from_qasm(self.to_qasm())
        return zxqc

    @classmethod
    def from_pyzx(cls, zxqc):
        """
        get pyzx's Circuit instance.

        Parameters
        ----------
        zxqc: instance of pyzx's Circuit
            quantum circuit

        Returns
        -------
        qc: instance of QCirc
            quantum circuit

        Notes
        -----
        Non-unitary gates: measure, reset are not supported.

        """
        import pyzx as zx
        from pyzx import Circuit

        qc = __class__.from_qasm(zxqc.to_qasm())
        return qc
        
    def optimize(self, *args, **kwargs):
        """
        optimize the quantum circuit (using pyzx's full_optimize method).

        Parameters
        ----------
        None

        Returns
        -------
        qc: instance of QCirc
            optimized circuit

        Notes
        -----
        Non-unitary gates: measure, reset are not supported.

        """
        import pyzx as zx
        from pyzx import Circuit

        zxqc = self.to_pyzx()
        zxqc_opt = zx.optimize.full_optimize(zxqc)
        qc = __class__.from_pyzx(zxqc_opt)

        return qc
        
    def equivalent(self, qc, *args, **kwargs):
        """
        two quantum circuits are equivalent or not (using pyzx's verify_equality method).

        Parameters
        ----------
        None

        Returns
        -------
        ans: bool

        Notes
        -----
        Non-unitary gates: measure, reset are not supported.

        """
        import pyzx as zx
        from pyzx import Circuit
        
        zxqc_A = Circuit.from_qasm(self.to_qasm())
        zxqc_B = Circuit.from_qasm(qc.to_qasm())
        ans =zxqc_A.verify_equality(zxqc_B)

        return ans

    def pop_gate(self):
        """
        pop first gate of the circuit.

        Parameters
        ----------
        None

        Returns
        -------
        gate : tupple of (int, [int,int], [float,float,float], int, int)
            tupple of (kind, qid, para, c, ctrl)
            - kind ... kind of gate
            - qid ... qubit id list
            - para ... parameters for rotation
            - c ... classical register ID to store measured data (only for measurement gate)
            - ctrl ... classical register id to controll the gate

        """
        (kind, qid, para, c, ctrl) = qcirc_pop_gate(self)
        return (kind, qid, para, c, ctrl)

    def append_gate(self, kind, qid, para=None, c=None, ctrl=None):
        """
        append gate to the end of the circuit.

        Parameters
        ----------
        kind : int
            kind of gate
        qid : list (int)
            list of qubit id
        para : list (float), default None
            list of parameters
        c : int, default None
            classical register id to store measured data
        ctrl : int, default None
            classical register id to controll the gate

        Returns
        -------
        None

        """
        qcirc_append_gate(self, kind, qid, para, c, ctrl)

    def split_unitary_non_unitary(self):
        """
        split two part of the gate.

        Parameters
        ----------
        None

        Returns
        -------
        qc_pair : tupple of (QCirc, Qcirc)
            former part includes only unitary gates and later part includes non-unitary gate (measure or reset) first
        """
        qc_unitary = __class__()
        qc_non_unitary = self.clone()
        while True:
            kind_ori = qc_non_unitary.kind_first()
            if kind_ori is None or kind_ori is MEASURE or kind_ori is RESET:
                break
            else:
                (kind, qid, para, c, ctrl) = qc_non_unitary.pop_gate()
                qc_unitary.append_gate(kind, qid, para, c, ctrl)

        qc_pair = (qc_unitary, qc_non_unitary)
        return qc_pair

    def __del__(self):
        
        qcirc_free(self)

    # non-unitary gate

    def measure(self, qid, cid):
        """
        add measurement gate (Z-basis).

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        cid : list of int
            classical register id list to store measured result.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=MEASURE, qid=qid, cid=cid, ctrl=None)
        return self

    def reset(self, qid):
        """
        add reset gate.

        Parameters
        ----------
        qid : list of int
            qubit id list to measure.
        cid : list of int
            classical register id list to store measured result.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        Notes
        -----
        'cid' must be 'None' or same length as 'qid'

        """
        self.__add_quantum_gate(kind=RESET, qid=qid, cid=None, ctrl=None)
        return self

    # add 1-qubit gate
    
    def x(self, q0, ctrl=None):
        """
        add X gate.

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=PAULI_X, qid=[q0], ctrl=ctrl)
        return self

    def y(self, q0, ctrl=None):
        """
        add Y gate.

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=PAULI_Y, qid=[q0], ctrl=ctrl)
        return self

    def z(self, q0, ctrl=None):
        """
        add Z gate.

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=PAULI_Z, qid=[q0], ctrl=ctrl)
        return self

    def h(self, q0, ctrl=None):
        """
        add H gate (hadamard gate).

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=HADAMARD, qid=[q0], ctrl=ctrl)
        return self
        
    def xr(self, q0, ctrl=None):
        """
        add root X gate.

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=ROOT_PAULI_X, qid=[q0], ctrl=ctrl)
        return self

    def xr_dg(self, q0, ctrl=None):
        """
        add root X dagger gate 
        (hermmitian conjugate of root X gate).

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=ROOT_PAULI_X_, qid=[q0], ctrl=ctrl)
        return self

    def s(self, q0, ctrl=None):
        """
        add S gate.

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=PHASE_SHIFT_S, qid=[q0], ctrl=ctrl)
        return self

    def s_dg(self, q0, ctrl=None):
        """
        add S dagger gate (hermitian conjugate of S gate).

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=PHASE_SHIFT_S_, qid=[q0], ctrl=ctrl)
        return self

    def t(self, q0, ctrl=None):
        """
        add T gate.

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=PHASE_SHIFT_T, qid=[q0], ctrl=ctrl)
        return self

    def t_dg(self, q0, ctrl=None):
        """
        add T dagger gate (hermitian conjugate of T gate).

        Parameters
        ----------
        q0 : int
            qubit id.
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=PHASE_SHIFT_T_, qid=[q0], ctrl=ctrl)
        return self

    def rx(self, q0, phase=DEF_PHASE, ctrl=None):
        """
        add RX gate (rotation around X-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=ROTATION_X, qid=[q0], phase=phase, ctrl=ctrl)
        return self

    def ry(self, q0, phase=DEF_PHASE, ctrl=None):
        """
        add RY gate (rotation around Y-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=ROTATION_Y, qid=[q0], phase=phase, ctrl=ctrl)
        return self

    def rz(self, q0, phase=DEF_PHASE, ctrl=None):
        """
        add RZ gate (rotation around Z-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=ROTATION_Z, qid=[q0], phase=phase, ctrl=ctrl)
        return self

    def p(self, q0, phase=DEF_PHASE, ctrl=None):
        """
        add P gate (phase shift gate).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        Notes
        -----
        matrix expression is following...
        | 1.0 0.0             |
        | 0.0 exp(i*phase*PI) |

        """
        self.__add_quantum_gate(kind=PHASE_SHIFT, qid=[q0], phase=phase, ctrl=ctrl)
        return self

    # def u1(self, q0, alpha=DEF_PHASE, ctrl=None):
    #     """
    #     add U1 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     ctrl : int
    #         address of classical memory to control gate operation.
    # 
    #     Returns
    #     -------
    #     self : instance of QCirc
    #         quantum circuit after adding
    # 
    #     Notes
    #     -----
    #     this opration is equal to P gate (phase shift gate)
    # 
    #     """
    #     self.__add_quantum_gate(kind=ROTATION_U1, qid=[q0], phase=alpha, ctrl=ctrl)
    #     return self
    # 
    # def u2(self, q0, alpha=DEF_PHASE, beta=DEF_PHASE, ctrl=None):
    #     """
    #     add U2 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is pi radian).
    #     beta : float
    #         rotation angle (unit of angle is pi radian).
    #     ctrl : int
    #         address of classical memory to control gate operation.
    # 
    #     Returns
    #     -------
    #     self : instance of QCirc
    #         quantum circuit after adding
    # 
    #     Notes
    #     -----
    #     matrix experssion is following...
    #     | 1/sqrt(2)              -exp(i*alpha*PI)/sqrt(2)       |
    #     | exp(i*beta*PI)/sqrt(2) exp(i*(alpha+beta)*PI)/sqrt(2) |
    # 
    #     """
    #     self.__add_quantum_gate(kind=ROTATION_U2, qid=[q0], phase=alpha, phase1=beta, ctrl=ctrl)
    #     return self
    # 
    # def u3(self, q0, alpha=DEF_PHASE, beta=DEF_PHASE, gamma=DEF_PHASE, ctrl=None):
    #     """
    #     add U3 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is pi radian).
    #     beta : float
    #         rotation angle (unit of angle is pi radian).
    #     gamma : float
    #         rotation angle (unit of angle is pi radian).
    #     ctrl : int
    #         address of classical memory to control gate operation.
    # 
    #     Returns
    #     -------
    #     self : instance of QCirc
    #         quantum circuit after adding
    # 
    #     Notes
    #     -----
    #     matrix expression is following...
    #     | cos(gamma/2)                -exp(i*alpha*PI)*sin(gamma/2)       |
    #     | exp(i*beta*PI)*sin(gamma/2) exp(i*(alpha+beta)*PI)*cos(gamma/2) |
    # 
    # 
    #     """
    #     self.__add_quantum_gate(kind=ROTATION_U3, qid=[q0], phase=alpha, phase1=beta,
    #                             phase2=gamma, ctrl=ctrl)
    #     return self

    # add 2-qubit gate
    
    def cx(self, q0, q1, ctrl=None):
        """
        add CX gate (controlled X gate, controlled NOT gate, CNOT gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_X, qid=[q0,q1], ctrl=ctrl)
        return self

    def cy(self, q0, q1, ctrl=None):
        """
        operate CY gate (controlled X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_Z, qid=[q0,q1], ctrl=ctrl)
        self.__add_quantum_gate(kind=CONTROLLED_X, qid=[q0,q1], ctrl=ctrl)
        self.__add_quantum_gate(kind=PHASE_SHIFT_S, qid=[q0], ctrl=ctrl)
        return self

    def cz(self, q0, q1, ctrl=None):
        """
        add CZ gate (controlled Z gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_Z, qid=[q0,q1], ctrl=ctrl)
        return self

    def cxr(self, q0, q1, ctrl=None):
        """
        add CXR gate (controlled root X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_XR, qid=[q0,q1], ctrl=ctrl)
        return self

    def cxr_dg(self, q0, q1, ctrl=None):
        """
        add CXR dagger gate (controlled XR dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_XR_, qid=[q0,q1], ctrl=ctrl)
        return self

    def ch(self, q0, q1, ctrl=None):
        """
        add CH gate (controlled H gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_H, qid=[q0,q1], ctrl=ctrl)
        return self

    def cs(self, q0, q1, ctrl=None):
        """
        add CS gate (controlled S gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_S, qid=[q0,q1], ctrl=ctrl)
        return self

    def cs_dg(self, q0, q1, ctrl=None):
        """
        add CS dagger gate (controlled S dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_S_, qid=[q0,q1], ctrl=ctrl)
        return self

    def ct(self, q0, q1, ctrl=None):
        """
        add CT gate (controlled T gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_T, qid=[q0,q1], ctrl=ctrl)
        return self

    def ct_dg(self, q0, q1, ctrl=None):
        """
        add CT dagger gate (controlled T dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_T_, qid=[q0,q1], ctrl=ctrl)
        return self

    def sw(self, q0, q1, ctrl=None):
        """
        add swap gate

        Parameters
        ----------
        q0 : int
            qubit id
        q1 : int
            qubit id
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=SWAP_QUBITS, qid=[q0,q1], ctrl=ctrl)
        return self

    def cp(self, q0, q1, phase=DEF_PHASE, ctrl=None):
        """
        add CP gate (controlled P gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_P, qid=[q0,q1], phase=phase, ctrl=ctrl)
        return self

    def crx(self, q0, q1, phase=DEF_PHASE, ctrl=None):
        """
        add CRX gate (controlled RX gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_RX, qid=[q0,q1], phase=phase, ctrl=ctrl)
        return self

    def cry(self, q0, q1, phase=DEF_PHASE, ctrl=None):
        """
        add CRY gate (controlled RY gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_RY, qid=[q0,q1], phase=phase, ctrl=ctrl)
        return self

    def crz(self, q0, q1, phase=DEF_PHASE, ctrl=None):
        """
        add CRZ gate (controlled RZ gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.__add_quantum_gate(kind=CONTROLLED_RZ, qid=[q0,q1], phase=phase, ctrl=ctrl)
        return self

    # def cu1(self, q0, q1, alpha=DEF_PHASE, ctrl=None):
    #     """
    #     add CU1 gate (controlled U1 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     ctrl : int
    #         address of classical memory to control gate operation.
    # 
    #     Returns
    #     -------
    #     self : instance of QCirc
    #         quantum circuit after adding
    # 
    #     """
    #     self.__add_quantum_gate(kind=CONTROLLED_U1, qid=[q0,q1], phase=alpha, ctrl=ctrl)
    #     return self
    # 
    # def cu2(self, q0, q1, alpha=DEF_PHASE, beta=DEF_PHASE, ctrl=None):
    #     """
    #     add CU2 gate (controlled U2 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    #     ctrl : int
    #         address of classical memory to control gate operation.
    # 
    #     Returns
    #     -------
    #     self : instance of QCirc
    #         quantum circuit after adding
    # 
    #     """
    #     self.__add_quantum_gate(kind=CONTROLLED_U2, qid=[q0,q1], phase=alpha, phase1=beta, ctrl=ctrl)
    #     return self
    # 
    # def cu3(self, q0, q1, alpha=DEF_PHASE, beta=DEF_PHASE, gamma=DEF_PHASE, ctrl=None):
    #     """
    #     add CU3 gate (controlled U3 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    #     gamma : float
    #         rotation angle (unit of angle is PI radian).
    #     ctrl : int
    #         address of classical memory to control gate operation.
    # 
    #     Returns
    #     -------
    #     self : instance of QCirc
    #         quantum circuit after adding
    # 
    #     """
    #     self.__add_quantum_gate(kind=CONTROLLED_U3, qid=[q0,q1], phase=alpha, phase1=beta,
    #                             phase2=gamma, ctrl=ctrl)
    #     return self

    # 3-qubit gate
    
    def ccx(self, q0, q1, q2, ctrl=None):
        """
        add CCX gate (toffoli gate, controlled controlled X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (control qubit).
        q2 : int
            qubit id (target qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.cxr(q1,q2,ctrl=ctrl).cx(q0,q1,ctrl=ctrl).cxr_dg(q1,q2,ctrl=ctrl)
        self.cx(q0,q1,ctrl=ctrl).cxr(q0,q2,ctrl=ctrl)
        return self

    def csw(self, q0, q1, q2, ctrl=None):
        """
        add CSW gate (fredkin gate, controlled swap gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (swap qubit).
        q2 : int
            qubit id (swap qubit).
        ctrl : int
            address of classical memory to control gate operation.

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        self.cx(q2,q1,ctrl=ctrl).ccx(q0,q1,q2,ctrl=ctrl).cx(q2,q1,ctrl=ctrl)
        return self

    def operate(self, pp=None, ctrl=None):
        """
        add unitary operator.

        Parameters
        ----------
        pp : instance of PauliProduct
            pauli product to operate
        ctrl : int
            contoroll qubit id for controlled pauli product

        Returns
        -------
        self : instance of QCirc
            quantum circuit after adding

        """
        pauli_list = pp.pauli_list
        qid = pp.qid
    
        if ctrl is None:
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.x(q)
                elif pauli == 'Y':
                    self.y(q)
                elif pauli == 'Z':
                    self.z(q)
                else:
                    continue
        else:
            if ctrl in qid:
                raise ValueError("controll and target qubit id conflict")
        
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.cx(ctrl, q)
                elif pauli == 'Y':
                    self.cy(ctrl, q)
                elif pauli == 'Z':
                    self.cz(ctrl, q)
                else:
                    continue
    
        return self

    def __add_quantum_gate(self, kind=None, qid=None, cid=None,
                           phase=DEF_PHASE, phase1=DEF_PHASE, phase2=DEF_PHASE,
                           ctrl=None):

        if qid is None:
            raise ValueError("qid must be specified")
            
        para = [phase, phase1, phase2]
        if is_measurement_gate(kind) == True:
            if cid is None:
                raise ValueError("cid must be specified")
            for q, c in zip(qid, cid):
                qid = [q]
                self.append_gate(kind, qid, para, c, ctrl)

        elif is_reset_gate(kind) == True:
            for q in qid:
                qid = [q]
                c = None
                self.append_gate(kind, qid, para, c, ctrl)
        
        else:
            c = None
            if kind == PAULI_Y:
                self.append_gate(PAULI_Z, qid, para, c, ctrl)
                self.append_gate(PAULI_X, qid, para, c, ctrl)
            elif kind == ROOT_PAULI_X:
                para[0] = 0.5
                self.append_gate(ROTATION_X, qid, para, c, ctrl)
            elif kind == ROOT_PAULI_X_:
                para[0] = -0.5
                self.append_gate(ROTATION_X, qid, para, c, ctrl)
            elif kind == PHASE_SHIFT:
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == ROTATION_Y:
                self.append_gate(PHASE_SHIFT_S_, qid, para, c, ctrl)
                self.append_gate(ROTATION_X, qid, para, c, ctrl)
                self.append_gate(PHASE_SHIFT_S, qid, para, c, ctrl)
            elif kind == CONTROLLED_XR:
                para[0] = 0.5
                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)
                para[0] = 0.25
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == CONTROLLED_XR_:
                para[0] = -0.5
                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)
                para[0] = -0.25
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == CONTROLLED_RX:
                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)
            elif kind == CONTROLLED_RY:
                # cs_dg gate
                phase = [-0.5, 0.0, 0.0]
                self.append_gate(CONTROLLED_RZ, qid, phase, c, ctrl)
                phase[0] = -0.25
                self.append_gate(ROTATION_Z, qid, phase, c, ctrl)

                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_H, qid, para, c, ctrl)

                # cs gate
                phase[0] = 0.5
                self.append_gate(CONTROLLED_RZ, qid, phase, c, ctrl)
                phase[0] = 0.25
                self.append_gate(ROTATION_Z, qid, phase, c, ctrl)
            elif kind == CONTROLLED_P:
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                para[0] = para[0] / 2
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == CONTROLLED_S:
                para[0] = 0.5
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                para[0] = 0.25
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == CONTROLLED_S_:
                para[0] = -0.5
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                para[0] = -0.25
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == CONTROLLED_T:
                para[0] = 0.25
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                para[0] = 0.125
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == CONTROLLED_T_:
                para[0] = -0.25
                self.append_gate(CONTROLLED_RZ, qid, para, c, ctrl)
                para[0] = -0.125
                self.append_gate(ROTATION_Z, qid, para, c, ctrl)
            elif kind == SWAP_QUBITS:
                qid_inv = qid[:]
                qid_inv[0], qid_inv[1] = qid[1], qid[0]
                self.append_gate(CONTROLLED_X, qid, para, c, ctrl)
                self.append_gate(CONTROLLED_X, qid_inv, para, c, ctrl)
                self.append_gate(CONTROLLED_X, qid, para, c, ctrl)
            else:
                self.append_gate(kind, qid, para, c, ctrl)
        
# c-library for qstate
from qlazy.lib.qcirc_c import *



# -*- coding: utf-8 -*-
import os
import pathlib
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext as build_ext_orig

_VERSION = '0.2.4'

class CMakeExtension(Extension):

    def __init__(self, name):
        super().__init__(name, sources=[])

class build_ext(build_ext_orig):

    def run(self):
        for ext in self.extensions:
            self.build_cmake(ext)
        super().run()

    def build_cmake(self, ext):
        cwd = pathlib.Path().absolute()
        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)
        extdir = pathlib.Path(self.get_ext_fullpath(ext.name))
        extdir.mkdir(parents=True, exist_ok=True)
        config = 'Debug' if self.debug else 'Release'
        cmake_args = [
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + str(extdir.parent.parent.absolute()),
            '-DCMAKE_RUNTIME_OUTPUT_DIRECTORY=' + str(extdir.parent.parent.parent.absolute()),
            '-DCMAKE_BUILD_TYPE=' + config
        ]
        build_args = [
            '--config', config,
            '--', '-j4'
        ]
        os.chdir(str(build_temp))
        self.spawn(['cmake', str(cwd)] + cmake_args)
        if not self.dry_run:
            self.spawn(['cmake', '--build', '.'] + build_args)
        os.chdir(str(cwd))

setup(
    name='qlazy',
    version=_VERSION,
    url='https://github.com/samn33/qlazy',
    author='Sam.N',
    author_email='saminriver33@gmail.com',
    description='Quantum Computing Simulator',
    long_description='',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy>=1.22.0',
        'pyzx>=0.7.0',
    ],
    license='Apache Software License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'License :: OSI Approved :: Apache Software License',
    ],
    keywords=['quantum', 'simulator'],
    ext_modules=[CMakeExtension('qlazy/lib/c/qlz')],
    cmdclass={
        'build_ext': build_ext,
    },
    entry_points="""
    [console_scripts]
    qlazy = qlazy.core:main
    """,
)



# -*- coding: utf-8 -*-
import ctypes
from ctypes.util import find_library
import pathlib

from qlazy.Observable import Observable
from qlazy.error import *
from qlazy.config import *
from qlazy.util import *

lib= ctypes.CDLL(str(pathlib.Path(__file__).with_name('libqlz.'+get_lib_ext())))
libc = ctypes.CDLL(find_library("c"),mode=ctypes.RTLD_GLOBAL)

def observable_init(str=None):
    
    c_str = ctypes.create_string_buffer(str.encode('utf-8'))

    observ = None
    c_observ = ctypes.c_void_p(observ)
        
    lib.observable_init.restype = ctypes.c_int
    lib.observable_init.argtypes = [ctypes.POINTER(ctypes.c_char),
                                    ctypes.POINTER(ctypes.c_void_p)]
    ret = lib.observable_init(c_str, c_observ)

    if ret == FALSE:
        raise Observable_Error_Initialize()
        
    return c_observ

def observable_free(ob):

    lib.observable_free.argtypes = [ctypes.POINTER(Observable)]
    lib.observable_free(ctypes.byref(ob))



# -*- coding: utf-8 -*-
import sys
import random
from collections import Counter
import numpy as np
import cmath

from qlazy.error import *
from qlazy.config import *
from qlazy.util import *
from qlazy.Result import *

from qulacs import QuantumState
from qulacs import QuantumCircuit
from qulacs.gate import Identity, X, Y, Z
from qulacs.gate import H, S, Sdag, T, Tdag, sqrtX, sqrtXdag, sqrtY, sqrtYdag
from qulacs.gate import CNOT, CZ, SWAP
from qulacs.gate import RX, RY, RZ
from qulacs.gate import U1, U2, U3
from qulacs.gate import Measurement
from qulacs.gate import DenseMatrix
from qulacs.gate import to_matrix_gate

GateFunctionName = {
    # 1-qubit, 0-parameter gate
    PAULI_X: 'X',
    PAULI_Y: 'Y',
    PAULI_Z: 'Z',
    ROOT_PAULI_X:'sqrtX',
    ROOT_PAULI_X_:'sqrtXdag',
    HADAMARD:'H',
    PHASE_SHIFT_S:'S',
    PHASE_SHIFT_S_:'Sdag',
    PHASE_SHIFT_T:'T',
    PHASE_SHIFT_T_:'Tdag',
    IDENTITY:'Identity',
    # 1-qubit, 1-parameter gate
    ROTATION_X:'RX',
    ROTATION_Y:'RY',
    ROTATION_Z:'RZ',
    PHASE_SHIFT:'__get_P',
    ROTATION_U1:'U1',
    # 1-qubit, 2-parameter gate
    ROTATION_U2:'U2',
    # 1-qubit, 3-parameter gate
    ROTATION_U3:'U3',
    # 2-qubit, 0-parameters gate
    CONTROLLED_X:'CNOT',
    CONTROLLED_Y:'__get_CY',
    CONTROLLED_Z:'CZ',
    CONTROLLED_XR:'__get_CXR',
    CONTROLLED_XR_:'__get_CXR_dg',
    CONTROLLED_H:'__get_CH',
    CONTROLLED_S:'__get_CS',
    CONTROLLED_S_:'__get_CS_dg',
    CONTROLLED_T:'__get_CT',
    CONTROLLED_T_:'__get_CT_dg',
    CONTROLLED_P:'__get_CP',
    SWAP_QUBITS:'SWAP',
    # 2-qubit, 1-parameters gate
    CONTROLLED_RX:'__get_CRX',
    CONTROLLED_RY:'__get_CRY',
    CONTROLLED_RZ:'__get_CRZ',
    CONTROLLED_U1:'__get_CU1',
    # 2-qubit, 2-parameters gate
    CONTROLLED_U2:'__get_CU2',
    # 2-qubit, 3-parameters gate
    CONTROLLED_U3:'__get_CU3',
}

def run_cpu(qcirc=[], shots=1, cid=[], backend=None):

    return __run_all(qcirc=qcirc, shots=shots, cid=cid, backend=backend, proc='CPU')

def run_gpu(qcirc=[], shots=1, cid=[], backend=None):

    return __run_all(qcirc=qcirc, shots=shots, cid=cid, backend=backend, proc='GPU')

def __run_all(qcirc=None, shots=1, cid=[], backend=None, proc='CPU'):

    if qcirc is None:
        raise ValueError("quantum circuit must be specified.")

    qubit_num = qcirc.qubit_num
    cmem_num = qcirc.cmem_num

    if cmem_num < len(cid):
        raise ValueError("length of cid must be less than classical resister size of qcirc")

    if cid == []:
        cid = [i for i in range(cmem_num)]

    #
    # initialize
    #

    if proc == 'CPU':
        qstate = QuantumState(qubit_num)
    else:
        qstate = QuantumStateGpu(qubit_num)
    cmem = [0] * cmem_num

    #
    # before measurement gate
    #

    while True:
        kind = qcirc.kind_first()
        # if kind == None or kind == MEASURE or kind == RESET:
        if kind is None or kind is MEASURE or kind is RESET:
            break

        else:
            (kind, qid, para, c, ctrl) = qcirc.pop_gate()
            if ctrl == None or (ctrl != None and cmem[ctrl] == 1):
                __qulacs_operate_qgate(qstate, qubit_num, kind=kind, qid=qid,
                                       phase=para[0], phase1=para[1], phase2=para[2])

    if kind == None:
        info = {'quantumstate': qstate, 'cmem': cmem}
        result = Result(cid=cid, frequency=None, backend=backend, info=info)
        return result
    
    #
    # after measurement gate
    #

    frequency = Counter()
    qstate_tmp = None
    for cnt in range(shots):

        qstate_tmp = qstate.copy()
        qcirc_tmp = qcirc.clone()

        while True:

            kind = qcirc_tmp.kind_first()
            if kind == None:
                break

            elif kind == MEASURE:
                (kind, qid, para, c, ctrl) = qcirc_tmp.pop_gate()
                mval = __qulacs_measure(qstate_tmp, qubit_num, qid[0])
                if c != None:
                    cmem[c] = mval
                        
            elif kind == RESET:
                (kind, qid, para, c, ctrl) = qcirc_tmp.pop_gate()
                __qulacs_reset(qstate_tmp, qubit_num, qid[0])

            else:
                (kind, qid, para, c, ctrl) = qcirc_tmp.pop_gate()
                if (ctrl == None or (ctrl != None and cmem[ctrl] == 1)):
                    __qulacs_operate_qgate(qstate_tmp, qubit_num, kind=kind, qid=qid,
                                           phase=para[0], phase1=para[1], phase2=para[2])

        if len(cmem) > 0:
            mval = ''.join(map(str, [cmem[i] for i in cid]))
            frequency[mval] += 1
    
    if qstate_tmp is not None:
        qstate.load(qstate_tmp.get_vector())

    if len(frequency) == 0:
        frequency = None

    info = {'quantumstate': qstate, 'cmem': cmem}
    result = Result(cid=cid, frequency=frequency, backend=backend, info=info)
    
    return result
        
def __is_supported_qgate(kind):

    if kind in GateFunctionName.keys():
        return True
    else:
        return False

# not supported as pre-defined gates

def __get_P(q0, phase):

    exp = cmath.exp(1.j * phase)
    gate = DenseMatrix(q0, [[1., 0.], [0., exp]])
    return gate

def __get_CY(q0, q1):

    gate = to_matrix_gate(Y(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CXR(q0, q1):

    gate = to_matrix_gate(sqrtX(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CXR_dg(q0, q1):

    gate = to_matrix_gate(sqrtXdag(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CH(q0, q1):

    gate = to_matrix_gate(H(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CS(q0, q1):

    gate = to_matrix_gate(S(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CS_dg(q0, q1):

    gate = to_matrix_gate(Sdag(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CT(q0, q1):

    gate = to_matrix_gate(T(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CT_dg(q0, q1):

    gate = to_matrix_gate(Tdag(q1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CP(q0, q1, phase):

    exp = cmath.exp(1.j * phase)
    gate = DenseMatrix(q1, [[1., 0.], [0., exp]])
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CRX(q0, q1, phase):

    gate = to_matrix_gate(RX(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CRY(q0, q1, phase):

    gate = to_matrix_gate(RY(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CRZ(q0, q1, phase):

    gate = to_matrix_gate(RZ(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CU1(q0, q1, phase):

    gate = to_matrix_gate(U1(q1, phase))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CU2(q0, q1, phase, phase1):

    gate = to_matrix_gate(U2(q1, phase, phase1))
    gate.add_control_qubit(q0, 1)
    return gate

def __get_CU3(q0, q1, phase, phase1, phase2):

    gate = to_matrix_gate(U3(q1, phase, phase1, phase2))
    gate.add_control_qubit(q0, 1)
    return gate

def __qulacs_operate_qgate(qstate, qubit_num, kind, qid, phase, phase1, phase2):

    if __is_supported_qgate(kind) == False:
        raise ValueError("not supported quantum gate")

    circ = QuantumCircuit(qubit_num)

    term_num = get_qgate_qubit_num(kind)
    para_num = get_qgate_param_num(kind)
    gate_function_name = GateFunctionName[kind]

    phase = phase * np.pi
    phase1 = phase1 * np.pi
    phase2 = phase2 * np.pi

    # the sign-definition of rotation gate on qulacs 
    if (kind == ROTATION_X or kind == ROTATION_Y or kind == ROTATION_Z or
        kind == CONTROLLED_RX or kind == CONTROLLED_RY or kind == CONTROLLED_RZ):
        phase = -phase
    # the argument-order-definition of U2 gate on qulacs 
    elif kind == ROTATION_U2 or kind == CONTROLLED_U2:
        phase1, phase = phase, phase1
    # the argument-order-definition of U3 gate on qulacs 
    elif kind == ROTATION_U3 or kind == CONTROLLED_U3:
        phase2, phase = phase, phase2
        
    if term_num == 1 and para_num == 0:
        circ.add_gate(eval(gate_function_name)(qid[0]))
    elif term_num == 1 and para_num == 1:
        circ.add_gate(eval(gate_function_name)(qid[0], phase))
    elif term_num == 1 and para_num == 2:
        circ.add_gate(eval(gate_function_name)(qid[0], phase, phase1))
    elif term_num == 1 and para_num == 3:
        circ.add_gate(eval(gate_function_name)(qid[0], phase, phase1, phase2))
    elif term_num == 2 and para_num == 0:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1]))
    elif term_num == 2 and para_num == 1:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1], phase))
    elif term_num == 2 and para_num == 2:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1], phase, phase1))
    elif term_num == 2 and para_num == 3:
        circ.add_gate(eval(gate_function_name)(qid[0], qid[1], phase, phase1, phase2))
    else:
        raise ValueError("not supported terminal or parameter-number")
    
    circ.update_quantum_state(qstate)

def __qulacs_reset(qstate, qubit_num, q):

    # error check
    if q >= qubit_num:
        raise ValueError("reset qubit id is out of bound")

    circ = QuantumCircuit(qubit_num)
    circ.add_gate(Measurement(q, 0))
    circ.update_quantum_state(qstate)
    circ_flip = QuantumCircuit(qubit_num)
    if qstate.get_classical_value(0) == 1:
        circ_flip.add_gate(X(q))
        circ_flip.update_quantum_state(qstate)

def __qulacs_measure(qstate, qubit_num, q):

    # error check
    if q >= qubit_num:
        raise ValueError("measurement qubit id is out of bound")

    circ = QuantumCircuit(qubit_num)
    circ.add_gate(Measurement(q, 0))
    circ.update_quantum_state(qstate)
    mval = qstate.get_classical_value(0)

    return mval



# -*- coding: utf-8 -*-
import random
import math
import numpy as np
import warnings
import ctypes

from qlazy.config import *
from qlazy.error import *
from qlazy.QState import *
from qlazy.lib.densop_mcx import *

class DensOp(ctypes.Structure):
    """ Density Operator

    Attributes
    ----------
    row : int
        dimension of density operator (= 2**qubit_num).
    col : int
        dimension of density operator (= 2**qubit_num).
    element : list of list of complex
        matrix elements of density operator.

    """

    _fields_ = [
        ('row', ctypes.c_int),
        ('col', ctypes.c_int),
        ('buf_id', ctypes.c_int),
        ('elm', ctypes.c_void_p),
        ('buffer_0', ctypes.c_void_p),
        ('buffer_1', ctypes.c_void_p),
        ('gbank', ctypes.c_void_p),
    ]
    
    def __new__(cls, qubit_num=0, qstate=[], prob=[], matrix=None, **kwargs):
        """
        Parameters
        ----------
        qubit_num : int
            qubit number of the density operator.
        qstate : list of instance of QState
            quantum states of the mixed state.
        prob : list of float
            probability of each quantum state.
        matrix : list of list of comprex
            matrix elements of the density operator.

        Examles
        -------
        >>> import numpy as np
        >>> qs_0 = QState(2)
        >>> qs_1 = QState(2).h(0).cx(0,1)
        >>> de_A = DensOp(qubit_num=5)  # = |00..0><00..0|
        >>> de_B = DensOp(qstate=[qs_0,qs_1], prob=[0.2,0.8])
        >>> de_C = DensOp(matrix=np.array([[0.7,0.0],[0.0,0.3]]))

        Notes
        -----
        You must set either 'qubit_num' or 'qstate' or 'matrix'.
        If 'prob' isn't set, equal probabilities are set.

        """
        if qstate != [] and prob == []:
            mixed_num = len(qstate)
            prob = [1.0/mixed_num for _ in range(mixed_num)]
        
        if qubit_num != 0:
            qstate = [QState(qubit_num=qubit_num)]
            prob = [1.0]
            obj = densop_init(qstate, prob)
        
        elif qstate != [] and prob != []:
            obj = densop_init(qstate, prob)

        else:
            obj = densop_init_with_matrix(matrix)

        self = ctypes.cast(obj.value, ctypes.POINTER(cls)).contents
        return self
    
    def __str__(self):

        return str(self.get_elm())

    def reset(self, qid=[]):
        """
        reset to |0><0| state.

        Parameters
        ----------
        qid : list
            qubit id's list to reset.

        Returns
        -------
        densop : instance of DensOp
            reset density operator.

        Notes
        -----
        If 'qid' is not set, whole system is reset.

        """
        densop_reset(self, qid=qid)
        return self
        
    @classmethod
    def mix(cls, densop=[], prob=[]):
        """
        linear sum of the density operators.

        Parameters
        ----------
        densop : list of instances of DensOp
            densitiy operators.
        prob : list of float
            probabilities (coefficients of the linear sum).

        """
        N = len(densop)
        
        if sum(prob) != 1.0:
            s = sum(prob)
            prob = [p/s for p in prob]
         
        de_out = densop[0].clone()
        de_out.mul(factor=prob[0])
        for i in range(1,len(densop)):
            de_tmp = densop[i].clone()
            de_tmp.mul(factor=prob[i])
            de_out.add(densop=de_tmp)

        return de_out

    @classmethod
    def add_method(cls, method):
        """
        add method (custum gate).

        Parameters
        ----------
        method : func
            method (custum gate) to add.

        Examples
        --------
        >>> def bell(self, q0, q1):
        >>>     self.h(q0).cx(q0,q1)
        >>> ...
        >>> DensOp.add_method(bell)
        >>> qs = DensOp(qubit_num=2)
        >>> qs.bell(0,1)
        >>> ...
        
        """
        setattr(cls, method.__name__, method)
        
    @classmethod
    def add_methods(cls, *methods):
        """
        add methods (custum gates).

        Parameters
        ----------
        methods : func, func, ...
            arguments of methods (custum gates) to add.

        Examples
        --------
        >>> def bell(self, q0, q1):
        >>>     self.h(q0).cx(q0,q1)
        >>>     return self
        >>> ...
        >>> def flip(self, q0, q1):
        >>>     self.x(q0).x(q1)
        >>>     return self
        >>> ...
        >>> DensOp.add_methods(bell, flip, ...)
        >>> de = DensOp(qubit=2)
        >>> de.bell(0,1)
        >>> de.flip(0,1)
        >>> ...
        
        """
        for method in methods:
            if callable(method):
                setattr(cls, method.__name__, method)
            else:
                raise DensOp_Error_AddMethods()
            
    @classmethod
    def create_register(cls, num):
        """
        create registers (qubit id's list) and initialize zero.

        Parameters
        ----------
        num : int
            qubit number you want to use.

        Returns
        -------
        qid : list of int
            qubit id's list.

        Examples
        --------
        >>> qid = DensOp.create_register(3)
        >>> print(qid)
        [0,0,0]
        
        """
        qid = [0]*num
        return qid

    @classmethod
    def init_register(cls, *args):
        """
        initialize registers (qubit id's list).

        Parameters
        ----------
        args : list, list,...
            arguments of qubit registers.

        Returns
        -------
        idx : int
            total qubit number.

        Examples
        --------
        >>> qid_0 = DensOp.create_register(3)
        >>> qid_1 = DensOp.create_register(2)
        >>> print(qid_0, qid_1)
        [0,0,0] [0,0]
        >>> qnum = DensOp.init_register(qid_0, qid_1)
        >>> print(qnum, qid_0, qid_1)
        5 [0,1,2] [3,4]

        """
        idx = 0
        for i in range(len(args)):
            for j in range(len(args[i])):
                args[i][j] = idx
                idx += 1
        return idx
    
    @classmethod
    def free_all(cls, *densops):
        """
        free memory of the all density operators.

        Parameters
        ----------
        densops : instance of DensOp,instance of DensOp,...
            set of DensOp instances

        Returns
        -------
        None

        """
        warnings.warn("No need to call 'free_all' method because free automatically, or you can use class method 'del_all' to free memory explicitly.")

    @classmethod
    def del_all(cls, *densops):
        """
        free memory of the all density operators.

        Parameters
        ----------
        densops : instance of DensOp,instance of DensOp,...
            set of DensOp instances

        Returns
        -------
        None

        """
        for de in densops:
            if type(de) is list or type(de) is tuple:
                cls.del_all(*de)
            elif type(de) is DensOp:
                del de
            else:
                raise DensOp_Error_FreeAll()

    @property
    def element(self):
        """ matrix elements of density operator. """
        return self.get_elm()
    
    def get_elm(self, qid=[]):
        """
        get the matrix elements of density operator.

        Parameters
        ----------
        qid : list of int, default - list of all of the qubit id
            qubit id's list.

        Returns
        -------
        elm : list of complex
            elements of the density matrix.

        Notes
        -----
        If 'qid' is set, matrix elements of specified density operator
        are got after partial trace for remaining qubits. If no qubits
        are specified, you get all matrix elements of the density
        operator.

        """
        de_part = self.partial(qid=qid)
        elm = densop_get_elm(de_part)
        return elm
    
    def show(self, qid=[], nonzero=False):
        """
        show the elements of density operator.

        Parameters
        ----------
        qid : list of int, default - list of all of the qubit id
            qubit id's list to show.
        nonzero : bool, default False
            if True, only non-zero elements are printed.

        Returns
        -------
        None

        Notes
        -----
        If 'qid' is set, it shows the matrix elements of density
        operator for partial quantum system specified by 'qid'. If
        'qid' isn't set, it shows the matrix elements of whole quantum
        system.

        Examples
        --------
        >>> de = DensOp(qubit_num=1)
        >>> ...
        >>> de.show()
        elm[0][0] = +0.5000+0.0000*i : 0.2500 |++++
        elm[0][1] = +0.0000+0.0000*i : 0.0000 |
        elm[1][0] = +0.0000+0.0000*i : 0.0000 |
        elm[1][1] = +0.5000+0.0000*i : 0.2500 |++++
        ...
        >>> de.show(nonzero=True)
        elm[0][0] = +0.5000+0.0000*i : 0.2500 |++++
        elm[1][1] = +0.5000+0.0000*i : 0.2500 |++++

        """
        de_part = self.partial(qid=qid)
        densop_print(de_part, nonzero)

    def clone(self):
        """
        get the copy of density operator.

        Parameters
        ----------
        None

        Returns
        -------
        densop : instance of DensOp
            copy of the original density operator.

        """
        obj = densop_copy(self)
        de = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return de
    
    def add(self, densop=None):
        """
        add the density operator.

        Parameters
        ----------
        densop : instance of DensOp
            density operator to add.

        Returns
        -------
        self : instance of DensOp

        Notes
        -----
        This method change the original density operator.

        """
        densop_add(self, densop=densop)
        return self

    def mul(self, factor=1.0):
        """
        multiply the density operator by factor.

        Parameters
        ----------
        factor : float, default 1.0
            number to multiply.

        Returns
        -------
        self : instance of DensOp

        Notes
        -----
        This method change the original density operator.

        """
        densop_mul(self, factor=factor)
        return self

    def trace(self):
        """
        get the trace of density operator.

        Parameters
        ----------
        None

        Returns
        -------
        trace : float
            trace of the density operator.

        """
        trace = densop_trace(self)
        return trace
        
    def sqtrace(self):
        """
        get the square trace of density operator.

        Parameters
        ----------
        None

        Returns
        -------
        sqtrace : float
            square trace of the density operator.

        """
        sqtrace = densop_sqtrace(self)
        return sqtrace
        
    def patrace(self, qid=[]):
        """
        get the partial trace of density operator.

        Parameters
        ----------
        qid : list of int
            qubit id's list to show.

        Returns
        -------
        densop : instance of DensOp
            density operator after partial trace.

        """
        obj = densop_patrace(self, qid=qid)
        de = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return de

    def partial(self, qid=[]):
        """
        get the density operator for partial system.

        Parameters
        ----------
        qid : list of int
            qubit id's list to show.

        Returns
        -------
        densop : instance of DensOp
            density operator for partial system 
            (= partial trace for remaining system).

        """
        if qid is None or qid == []:
            return self.clone()
        else:
            qubit_num = int(math.log2(self.row))
            qid_remained = []
            for x in range(qubit_num):
                if not x in qid:
                    qid_remained.append(x)
            de_remained = self.patrace(qid=qid_remained)
            return de_remained
        
    def tenspro(self, densop):
        """
        get the tensor product with density operator.

        Parameters
        ----------
        densop : instance of DensOp
            density operator to get the tensor product..

        Returns
        -------
        densop_out : instance of DensOp
            tensor produt of 'self' and 'densop'.

        """
        obj = densop_tensor_product(self, densop)
        de = ctypes.cast(obj.value, ctypes.POINTER(self.__class__)).contents
        return de

    def composite(self, num=0):
        """
        get the composite density operator of same density operators.

        Parameters
        ----------
        num : int
            number of density operators..

        Returns
        -------
        de : instance of DensOp
            composite density operator.

        """
        if num <= 1:
            return self
        else:
            de = self.clone()
            for i in range(num-1):
                de_tmp = de.tenspro(self)
                de = de_tmp.clone()
            return de
        
    def join(self, de_list):
        """
        get tensor product state (density operator) of the density operators' list.

        Parameters
        ----------
        de_list : list (DensOp)
            list of density operators.

        Returns
        -------
        de_out : instance of DensOp
            tensor product state (density operator).

        """
        de_out = self.clone()
        for de in de_list:
            de_tmp = de_out.clone()
            de_out = de_tmp.tenspro(de)
        return de_out

    def expect(self, matrix=None):
        """
        get the expectation value of matrix under this density operator.

        Parameters
        ----------
        matrix : list of list of complex
            matrix expression of hermitian operator.

        Returns
        -------
        value : float
            expectation value.

        Notes
        -----
        'matrix' must be hermitian, and its dimension is equal to
        the dimension of density operator.

        """
        densop = self.clone()
        densop_apply_matrix(densop, matrix=matrix, dire='left')
        value = densop.trace()
        return value
        
    def apply(self, matrix=None, qid=[], dire='both'):
        """
        apply the matrix to density operator.
        (= [matrix] * [self] * [dagger of matrix])

        Parameters
        ----------
        matrix : list of list
            matrix to apply.

        Returns
        -------
        None

        Notes
        -----
        If 'qid' isn't set, dimension of the matrix must be equal to
        dimension of the density operator. If 'qid' is set, dimension
        of the matrix must be equal to the 2 power of 'qid' length.

        """
        densop_apply_matrix(self, matrix=matrix, qid=qid, dire=dire)
        return self

    def probability(self, kraus=[], povm=[], qid=[]):
        """
        get the probabilities for measuring operators. 
        (Kraus or POVM operators).

        Parameters
        ----------
        kraus : list of list of comprex
            Kraus operators.
        povm : list of list of comprex
            POVM operators.
        qid : list
            qubit id's list to measure.

        Returns
        -------
        prob : list of float
            probabilities for measuring operators.

        Notes
        -----
        Either 'kraus' or 'povm' must be set. If 'qid' is not set, all
        of the qubits are measured, the dimention of Kraus or POVM
        operator must be equal to the dimension of density
        operator. If 'qid' is set, the part of qubits are measured,
        the dimension of Kraus or POVM operator must be equal to the 2
        power of 'qid' length. This method does not change the
        original density operator.

        """
        if kraus != []:
            N = len(kraus)
            prob = [0.0]*N
            for i in range(N):
                 prob[i] = densop_probability(self, matrix=kraus[i], qid=qid,
                                             matrix_type='kraus')
                 if abs(prob[i]) < EPS:
                     prob[i] = 0.0
        elif povm != []:
            N = len(povm)
            prob = [0.0]*N
            for i in range(N):
                prob[i] = densop_probability(self, matrix=povm[i], qid=qid,
                                             matrix_type='povm')
                if abs(prob[i]) < EPS:
                    prob[i] = 0.0
        else:
            raise DensOp_Error_Probability()
                
        return prob

    def instrument(self, kraus=[], qid=[], measured_value=None):
        """
        instrument to the density operator

        Parameters
        ----------
        kraus : list of list of comprex
            Kraus operators.
        qid : list
            qubit id's list to measure.
        measured_value : int
            index of measurement 
            (in the case of selective measurement).

        Returns
        -------
        self : instance of DensOp

        Notes
        -----
        If 'measured_value' is set, selective measurement of the index
        corresponding to 'measured_value' is done (but not
        normalize). If 'measured_value' is not set, non-selective
        measurement is done (only operator sum for the Kraus operators
        are executed). This method does change the original density
        operator.

        """
        if qid is None or qid == []:
            qnum = int(math.log2(self.row))
            qid = [i for i in range(qnum)]

        if kraus == []:
            raise DensOp_Error_Instrument()
        else:
            N = len(kraus)

        if measured_value is None:  # non-selective measurement
            
            densop_ori = self.clone()
            for i in range(N):
                if i == 0:
                    self.apply(matrix=kraus[i], qid=qid, dire='both')
                else:
                    densop_tmp = densop_ori.clone()
                    densop_tmp.apply(matrix=kraus[i], qid=qid, dire='both')
                    self.add(densop=densop_tmp)
                
        else:  # selective measurement

            if (measured_value < 0 or measured_value >= N):
                raise DensOp_Error_Instrument()
            self.apply(matrix=kraus[measured_value],qid=qid)
            #prob = self.trace()
            #self.mul(factor=1.0/prob)

        return self

    def bit_flip(self, q, prob=0.0):
        """
        execute the quantum channel of bit flip.

        Parameters
        ----------
        q : int
            qubit id to execute.
        prob : float
            probabillity of the quantum channel.

        Returns
        -------
        self : instance of DensOp

        """
        Sigma_0 = np.eye(2)
        Sigma_1 = np.array([[0,1],[1,0]])
        Sigma_2 = np.array([[0,-1j],[1j,0]])
        Sigma_3 = np.array([[1,0],[0,-1]])

        para_a = math.sqrt(1-prob)
        para_b = math.sqrt(prob)
        kraus = [para_a*Sigma_0, para_b*Sigma_1]
        self.instrument(qid=[q], kraus=kraus)
        return self
    
    def phase_flip(self, q, prob=0.0):
        """
        execute the quantum channel of phase flip.

        Parameters
        ----------
        q : int
            qubit id to execute.
        prob : float
            probability of the quantum channel.

        Returns
        -------
        self : instance of DensOp

        """
        Sigma_0 = np.eye(2)
        Sigma_1 = np.array([[0,1],[1,0]])
        Sigma_2 = np.array([[0,-1j],[1j,0]])
        Sigma_3 = np.array([[1,0],[0,-1]])

        para_a = math.sqrt(1-prob)
        para_b = math.sqrt(prob)
        kraus = [para_a*Sigma_0, para_b*Sigma_3]
        self.instrument(qid=[q], kraus=kraus)
        return self
    
    def bit_phase_flip(self, q, prob=0.0):
        """
        execute the quantum channel of bit and phase flip.

        Parameters
        ----------
        q : int
            qubit id to execute.
        prob : float
            probability of the quantum channel.

        Returns
        -------
        self : instance of DensOp

        """
        Sigma_0 = np.eye(2)
        Sigma_1 = np.array([[0,1],[1,0]])
        Sigma_2 = np.array([[0,-1j],[1j,0]])
        Sigma_3 = np.array([[1,0],[0,-1]])

        para_a = math.sqrt(1-prob)
        para_b = math.sqrt(prob)
        kraus = [para_a*Sigma_0, para_b*Sigma_2]
        self.instrument(qid=[q], kraus=kraus)
        return self
    
    def depolarize(self, q, prob=0.0):
        """
        execute the quantum channel of depolarize.

        Parameters
        ----------
        q : int
            qubit id to execute.
        prob : float
            probability of the quantum channel.

        Returns
        -------
        self : instance of DensOp

        """
        Sigma_0 = np.eye(2)
        Sigma_1 = np.array([[0,1],[1,0]])
        Sigma_2 = np.array([[0,-1j],[1j,0]])
        Sigma_3 = np.array([[1,0],[0,-1]])

        para_a = math.sqrt(1-0.75*prob)
        para_b = math.sqrt(0.25*prob)
        para_c = math.sqrt(0.25*prob)
        para_d = math.sqrt(0.25*prob)
        kraus = [para_a*Sigma_0, para_b*Sigma_1, para_c*Sigma_2, para_d*Sigma_3]
        self.instrument(qid=[q], kraus=kraus)
        return self
    
    def amp_dump(self, q, prob=0.0):
        """
        execute the quantum channel of amplitude dumping.

        Parameters
        ----------
        q : int
            qubit id to execute.
        prob : float
            probability of the quantum channel.

        Returns
        -------
        self : instance of DensOp

        """
        transmit = math.sqrt(1.0-prob)
        reflect = math.sqrt(prob)
        kraus = [np.array([[1,0],[0,transmit]]), np.array([[0,reflect],[0,0]])]
        self.instrument(qid=[q], kraus=kraus)
        return self
    
    def phase_dump(self, q, prob=0.0):
        """
        execute the quantum channel of phase dumping.

        Parameters
        ----------
        q : int
            qubit id to execute.
        prob : float
            probability of the quantum channel.

        Returns
        -------
        self : instance of DensOp

        """
        transmit = math.sqrt(1.0-prob)
        reflect = math.sqrt(prob)
        kraus = [np.array([[1,0],[0,transmit]]), np.array([[0,0],[0,reflect]])]
        self.instrument(qid=[q], kraus=kraus)
        return self
    
    def __mat_sqrt(self,mat):  # mat is hermite

        eigenvals, unitary = np.linalg.eigh(mat)
        unitary_dg = np.conjugate(unitary.T)
        mat_diag = np.sqrt(np.abs(np.diag(eigenvals)))
        mat_sq = np.dot(np.dot(unitary,mat_diag),unitary_dg)

        return mat_sq

    def __mat_norm(self,mat):

        mat2 = np.dot(np.conjugate(mat.T),mat)  # mat2 is hermite
        mat2_sqrt = self.__mat_sqrt(mat2)
        norm = np.trace(mat2_sqrt).real

        return norm 

    def fidelity(self, densop=None, qid=[]):
        """
        get the fidelity with density operator.

        Parameters
        ----------
        densop : instance of DensOp
            density operators to get fidelity.
        qid : list of int, default - all of the qubit id's list
            qubit id's list.

        Returns
        -------
        fid : float
            fidelity of two density operators.

        """
        mat1 = self.get_elm(qid=qid)
        mat2 = densop.get_elm(qid=qid)

        if mat1.shape != mat2.shape:
            raise DensOp_Error_Fidelity()

        mat1_sqrt = self.__mat_sqrt(mat1)
        mat2_sqrt = self.__mat_sqrt(mat2)

        fid = self.__mat_norm(np.dot(mat1_sqrt,mat2_sqrt))
            
        return fid

    def distance(self, densop=None, qid=[]):
        """
        get the trace distance with the density operator.

        Parameters
        ----------
        densop : instance of DensOp
            density operator to get the trace distance.

        qid : list of int, default - all of the qubit id's list
            qubit id's list.

        Returns
        -------
        dis : float
            trace distance.

        """
        mat1 = self.get_elm(qid=qid)
        mat2 = densop.get_elm(qid=qid)

        if mat1.shape != mat2.shape:
            raise DensOp_Error_Distance()

        dis = 0.5 * self.__mat_norm(mat1-mat2)
            
        return dis

    def __mat_spectrum(self, mat):  # mat is hermite

        eigenvals, unitary = np.linalg.eigh(mat)
        unitary_T = unitary.T
        return eigenvals,unitary_T

    def spectrum(self):
        """
        get the spectrum.

        Parameters
        ----------
        None

        Returns
        -------
        qstate : list of QState
            list of the quantum state basis.
        prob : list of float
            list of coefficients for each quantum states basis.

        """
        mat = self.get_elm()
        eigvals,eigvecs = self.__mat_spectrum(mat)
        prob = [eigvals[i] for i in range(len(eigvals)) if abs(eigvals[i]) > EPS]
        vecs = [eigvecs[i] for i in range(len(eigvals)) if abs(eigvals[i]) > EPS]
        qstate = [QState(vector=vecs[i]) for i in range(len(prob))]

        #return prob,qstate
        return qstate,prob

    def __von_neumann_entropy(self):  # von neumann entropy

        mat = self.get_elm()
        eigvals = np.linalg.eigvalsh(mat)
        diag = [-eigvals[i]*np.log2(eigvals[i])
                for i in range(len(eigvals)) if abs(eigvals[i]) > EPS]
        ent = np.sum(diag)
        return ent
    
    def entropy(self, qid=[]):
        """
        get the von neumann entropy (or entanglement entropy).

        Parameters
        ----------
        qid : list of int, default - list all of the qubit id
            qubit id's list (sub-system) to calculate the entropy.

        Returns
        -------
        ent : float
            von neumann entropy.

        """
        qubit_num = int(math.log2(self.row))
        
        if qid == []:
            ent = self.__von_neumann_entropy()
        else:
            if (min(qid) < 0 or max(qid) >= qubit_num or len(qid)!=len(set(qid))):
                raise DensOp_Error_Entropy()
            if len(qid) == qubit_num:
                ent = self.__von_neumann_entropy()
            else:
                de_part = self.partial(qid=qid)
                ent = de_part.__von_neumann_entropy()
                
        return ent

    def cond_entropy(self, qid_0=[], qid_1=[]):
        """
        get the conditional entropy.

        Parameters
        ----------
        qid_0 : list of int
            qubit id's list (sub-system) to calculate the entropy.
        qid_1 : list of int
            qubit id's list (sub-system) to calculate the entropy.

        Returns
        -------
        ent : float
            conditianal entropy (S(qid_0|qid_1)).
        
        """
        qubit_num = int(math.log2(self.row))
        
        if (qid_0 == [] or qid_1 == []
            or min(qid_0) < 0 or max(qid_0) >= qubit_num
            or min(qid_1) < 0 or max(qid_1) >= qubit_num
            or len(qid_0) != len(set(qid_0))
            or len(qid_1) != len(set(qid_1))):
            raise DensOp_Error_Entropy()
        else:
            qid_merge = qid_0 + qid_1
            qid_whole = set(qid_merge)
            ent = self.entropy(qid_whole) - self.entropy(qid_1)
            
        return ent

    def mutual_info(self, qid_0=[], qid_1=[]):  # mutual information
        """
        get the mutual information.

        Parameters
        ----------
        qid_0 : list of int
            qubit id's list (sub-system) to calculate the mutual information.
        qid_1 : list of int
            qubit id's list (sub-system) to calculate the mutual information.

        Returns
        -------
        ent : float
            mutual information (S(qid_0:qid_1)).
        
        """
        qubit_num = int(math.log2(self.row))
        
        if (qid_0 == [] or qid_1 == []
            or min(qid_0) < 0 or max(qid_0) >= qubit_num
            or min(qid_1) < 0 or max(qid_1) >= qubit_num
            or len(qid_0) != len(set(qid_0))
            or len(qid_1) != len(set(qid_1))):
            raise DensOp_Error_Entropy()
        else:
            ent = self.entropy(qid_0) - self.cond_entropy(qid_0,qid_1)
            
        return ent

    def relative_entropy(self, densop=None):  # relative entropy
        """
        get the relative entropy.

        Parameters
        ----------
        densop : instance of DensOp
            density operator to get the relative entropy.

        Returns
        -------
        ent : float
            relative entropy.
        
        """
        if self.row != densop.row:
            raise DensOp_Error_Entropy()
        
        mat_A = self.get_elm()
        mat_B = densop.get_elm()

        eigvals_A,eigvecs_A = self.__mat_spectrum(mat_A)
        eigvals_B,eigvecs_B = self.__mat_spectrum(mat_B)

        P = np.dot(np.conjugate(eigvecs_A.T),eigvecs_B)
        P = np.conjugate(P)*P
        
        diag_A = [eigvals_A[i]*np.log2(eigvals_A[i])
                for i in range(len(eigvals_A)) if abs(eigvals_A[i]) > EPS]
        relent_A = np.sum(diag_A)

        relent_B = 0.0
        for i in range(len(eigvals_A)):
            if eigvals_A[i] < EPS:
                continue
            for j in range(len(eigvals_B)):
                relent_B += abs(P[i][j]) * eigvals_A[i] *np.log2(eigvals_B[j])
        
        relent = relent_A - relent_B
        return relent

    # 1-qubit gate

    def x(self, q0):
        """
        operate X gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=PAULI_X, phase=DEF_PHASE, qid=[q0])
        return self

    def y(self, q0):
        """
        operate Y gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=PAULI_Y, phase=DEF_PHASE, qid=[q0])
        return self

    def z(self, q0):
        """
        operate Z gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=PAULI_Z, phase=DEF_PHASE, qid=[q0])
        return self

    def xr(self, q0):
        """
        operate root X gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self,kind=ROOT_PAULI_X, phase=DEF_PHASE, qid=[q0])
        return self

    def xr_dg(self, q0):
        """
        operate root X dagger gate 
        (hermmitian conjugate of root X gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=ROOT_PAULI_X_, phase=DEF_PHASE, qid=[q0])
        return self

    def h(self, q0):
        """
        operate H gate (hadamard gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=HADAMARD, phase=DEF_PHASE, qid=[q0])
        return self

    def s(self, q0):
        """
        operate S gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=PHASE_SHIFT_S, phase=DEF_PHASE, qid=[q0])
        return self

    def s_dg(self, q0):
        """
        operate S dagger gate (hermitian conjugate of S gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=PHASE_SHIFT_S_, phase=DEF_PHASE, qid=[q0])
        return self

    def t(self, q0):
        """
        operate T gate.

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=PHASE_SHIFT_T, phase=DEF_PHASE, qid=[q0])
        return self

    def t_dg(self, q0):
        """
        operate T dagger gate (hermitian conjugate of T gate).

        Parameters
        ----------
        q0 : int
            qubit id.

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=PHASE_SHIFT_T_, phase=DEF_PHASE, qid=[q0])
        return self

    def rx(self, q0, phase=DEF_PHASE):
        """
        operate RX gate (rotation around X-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=ROTATION_X, phase=phase, qid=[q0])
        return self

    def ry(self, q0, phase=DEF_PHASE):
        """
        operate RY gate (rotation around Y-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=ROTATION_Y, phase=phase, qid=[q0])
        return self

    def rz(self, q0, phase=DEF_PHASE):
        """
        operate RZ gate (rotation around Z-axis).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=ROTATION_Z, phase=phase, qid=[q0])
        return self

    def p(self, q0, phase=DEF_PHASE):
        """
        operate P gate (phase shift gate).

        Parameters
        ----------
        q0 : int
            qubit id.
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of DensOp

        Notes
        -----
        matrix expression is following...
        | 1.0 0.0             |
        | 0.0 exp(i*phase*PI) |

        """
        densop_operate_qgate(self, kind=PHASE_SHIFT, phase=phase, qid=[q0])
        return self

    # def u1(self, q0, alpha=DEF_PHASE):
    #     """
    #     operate U1 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of DensOp
    # 
    #     Notes
    #     -----
    #     this opration is equal to P gate (phase shift gate)
    # 
    #     """
    #     densop_operate_qgate(self, kind=ROTATION_U1, phase=alpha, qid=[q0])
    #     return self
    # 
    # def u2(self, q0, alpha=DEF_PHASE, beta=DEF_PHASE):
    #     """
    #     operate U2 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of DensOp
    # 
    #     Notes
    #     -----
    #     matrix experssion is following...
    #     | 1/sqrt(2)              -exp(i*alpha*PI)/sqrt(2)       |
    #     | exp(i*beta*PI)/sqrt(2) exp(i*(alpha+beta)*PI)/sqrt(2) |
    # 
    #     """
    #     densop_operate_qgate(self, kind=ROTATION_U2, phase=alpha, phase1=beta, qid=[q0])
    #     return self
    # 
    # def u3(self, q0, alpha=DEF_PHASE, beta=DEF_PHASE, gamma=DEF_PHASE):
    #     """
    #     operate U3 gate (by IBM).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id.
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    #     gamma : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of DensOp
    # 
    #     Notes
    #     -----
    #     matrix expression is following...
    #     | cos(gamma/2)                -exp(i*alpha*PI)*sin(gamma/2)       |
    #     | exp(i*beta*PI)*sin(gamma/2) exp(i*(alpha+beta)*PI)*cos(gamma/2) |
    # 
    # 
    #     """
    #     densop_operate_qgate(self, kind=ROTATION_U3, phase=alpha, phase1=beta,
    #                          phase2=gamma, qid=[q0])
    #     return self

    # 2-qubit gate

    def cx(self, q0, q1):
        """
        operate CX gate (controlled X gate, controlled NOT gate, CNOT gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_X, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cy(self, q0, q1):
        """
        operate CY gate (controlled X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_Y, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cz(self, q0, q1):
        """
        operate CZ gate (controlled Z gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_Z, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cxr(self, q0, q1):
        """
        operate CXR gate (controlled root X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_XR, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cxr_dg(self, q0, q1):
        """
        operate CXR dagger gate (controlled XR dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_XR_, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def ch(self, q0, q1):
        """
        operate CH gate (controlled H gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_H, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cs(self, q0, q1):
        """
        operate CS gate (controlled S gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_S, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cs_dg(self, q0, q1):
        """
        operate CS dagger gate (controlled S dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_S_, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def ct(self, q0, q1):
        """
        operate CT gate (controlled T gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_T, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def ct_dg(self, q0, q1):
        """
        operate CT dagger gate (controlled T dagger gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_T_, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def sw(self, q0, q1):
        """
        swap gate

        Parameters
        ----------
        q0 : int
            qubit id
        q1 : int
            qubit id

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=SWAP_QUBITS, phase=DEF_PHASE, qid=[q0,q1])
        return self

    def cp(self, q0, q1, phase=DEF_PHASE):
        """
        operate CP gate (controlled P gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_P, phase=phase, qid=[q0,q1])
        return self

    def crx(self, q0, q1, phase=DEF_PHASE):
        """
        operate CRX gate (controlled RX gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_RX, phase=phase, qid=[q0,q1])
        return self

    def cry(self, q0, q1, phase=DEF_PHASE):
        """
        operate CRY gate (controlled RY gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_RY, phase=phase, qid=[q0,q1])
        return self

    def crz(self, q0, q1, phase=DEF_PHASE):
        """
        operate CRZ gate (controlled RZ gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (target qubit).
        phase : float
            rotation angle (unit of angle is PI radian).

        Returns
        -------
        self : instance of DensOp

        """
        densop_operate_qgate(self, kind=CONTROLLED_RZ, phase=phase, qid=[q0,q1])
        return self

    # def cu1(self, q0, q1, alpha=DEF_PHASE):
    #     """
    #     operate CU1 gate (controlled U1 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of DensOp
    # 
    #     """
    #     densop_operate_qgate(self, kind=CONTROLLED_U1, phase=alpha, qid=[q0,q1])
    #     return self
    # 
    # def cu2(self, q0, q1, alpha=DEF_PHASE, beta=DEF_PHASE):
    #     """
    #     operate CU2 gate (controlled U2 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of DensOp
    # 
    #     """
    #     densop_operate_qgate(self, kind=CONTROLLED_U2, phase=alpha, phase1=beta,
    #                          qid=[q0,q1])
    #     return self
    # 
    # def cu3(self, q0, q1, alpha=DEF_PHASE, beta=DEF_PHASE, gamma=DEF_PHASE):
    #     """
    #     operate CU3 gate (controlled U3 gate).
    # 
    #     Parameters
    #     ----------
    #     q0 : int
    #         qubit id (control qubit).
    #     q1 : int
    #         qubit id (target qubit).
    #     alpha : float
    #         rotation angle (unit of angle is PI radian).
    #     beta : float
    #         rotation angle (unit of angle is PI radian).
    #     gamma : float
    #         rotation angle (unit of angle is PI radian).
    # 
    #     Returns
    #     -------
    #     self : instance of DensOp
    # 
    #     """
    #     densop_operate_qgate(self, kind=CONTROLLED_U3, phase=alpha, phase1=beta,
    #                          phase2=gamma, qid=[q0,q1])
    #     return self

    # 3-qubit gate
    
    def ccx(self, q0, q1, q2):
        """
        operate CCX gate (toffoli gate, controlled controlled X gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (control qubit).
        q2 : int
            qubit id (target qubit).

        Returns
        -------
        self : instance of DensOp

        """
        self.cxr(q1,q2).cx(q0,q1).cxr_dg(q1,q2).cx(q0,q1).cxr(q0,q2)
        return self

    def csw(self, q0, q1, q2):
        """
        operate CSW gate (fredkin gate, controlled swap gate).

        Parameters
        ----------
        q0 : int
            qubit id (control qubit).
        q1 : int
            qubit id (swap qubit).
        q2 : int
            qubit id (swap qubit).

        Returns
        -------
        self : instance of DensOp

        """
        self.cx(q2,q1).ccx(q0,q1,q2).cx(q2,q1)
        return self
    
    # other gate
    
    def mcx(self,qid=[]):
        """
        operate MCX gate (multi-controlled X gate).
       
        Parameters
        ----------
        qid : list of int
            qubit id list [control, control, ... , control, target]
    
        Returns
        -------
        self : instance of DensOp
        
        """
        densop_mcx(self, qid)
        return self

    def operate(self, pp=None, ctrl=None):
        """
        operate unitary operator to density operator.

        Parameters
        ----------
        pp : instance of PauliProduct
            pauli product to operate
        ctrl : int
            contoroll qubit id for controlled pauli product

        Returns
        -------
        self : instance of DensOp
            density operator after operation

        """
        pauli_list = pp.pauli_list
        qid = pp.qid
    
        if ctrl is None:
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.x(q)
                elif pauli == 'Y':
                    self.y(q)
                elif pauli == 'Z':
                    self.z(q)
                else:
                    continue
        else:
            if ctrl in qid:
                raise ValueError("controll and target qubit id conflict")
        
            for q, pauli in zip(qid, pauli_list):
                if pauli == 'X':
                    self.cx(ctrl, q)
                elif pauli == 'Y':
                    self.cy(ctrl, q)
                elif pauli == 'Z':
                    self.cz(ctrl, q)
                else:
                    continue
    
        return self

    def free(self):
        """
        free memory of quantum state.

        Parameters
        ----------
        None

        Returns
        -------
        None
        self : instance of DensOp

        """
        # densop_free(self)
        warnings.warn("No need to call 'free' method because free automatically, or you can use 'del' to free memory explicitly.")

    def __del__(self):

        densop_free(self)
    
# c-library for densop
from qlazy.lib.densop_c import *



#
# followings must be equal to definitions in 'qlazy.h'
#

TRUE  = 1
FALSE = 0

EPS = 1e-6
INF = 1e+6

MAX_QUBIT_NUM = 30

DEF_SHOTS = 1

DEF_TAG = 'DEFAULT'

DEF_PHASE  = 0.0
DEF_ANGLE  = 0.0

BELL_PHI_PLUS  = 0
BELL_PHI_MINUS = 3
BELL_PSI_PLUS  = 1
BELL_PSI_MINUS = 2

# Kind

CIRC           = 1
GATES          = 2
SHOW           = 3
BLOCH          = 4
ECHO           = 5
OUTPUT         = 6
HELP           = 7
QUIT	       = 8
INIT           = 9
PAULI_X        = 120
PAULI_Y        = 121
PAULI_Z        = 122
ROOT_PAULI_X   = 123
ROOT_PAULI_X_  = 124
HADAMARD       = 130
PHASE_SHIFT_S  = 140
PHASE_SHIFT_S_ = 141
PHASE_SHIFT_T  = 142
PHASE_SHIFT_T_ = 143
PHASE_SHIFT    = 144
ROTATION_X     = 150
ROTATION_Y     = 151
ROTATION_Z     = 152
ROTATION_U1    = 153
ROTATION_U2    = 154
ROTATION_U3    = 155
CONTROLLED_X   = 160
CONTROLLED_Y   = 161
CONTROLLED_Z   = 162
CONTROLLED_XR  = 163
CONTROLLED_XR_ = 164
CONTROLLED_H   = 165
CONTROLLED_S   = 166
CONTROLLED_S_  = 167
CONTROLLED_T   = 168
CONTROLLED_T_  = 169
CONTROLLED_P   = 170
CONTROLLED_RX  = 171
CONTROLLED_RY  = 172
CONTROLLED_RZ  = 173
CONTROLLED_U1  = 174
CONTROLLED_U2  = 175
CONTROLLED_U3  = 176
SWAP_QUBITS    = 180
MEASURE        = 200
MEASURE_X      = 201
MEASURE_Y      = 202
MEASURE_Z      = 203
MEASURE_BELL   = 204
RESET          = 205
NOT_A_GATE     = 1000
IDENTITY       = 2000

GATE_KIND = {
    'x': PAULI_X,
    'y': PAULI_Y,
    'z': PAULI_Z,
    'xr': ROOT_PAULI_X,
    'xr_dg': ROOT_PAULI_X_,
    'h': HADAMARD,
    's': PHASE_SHIFT_S,
    's_dg': PHASE_SHIFT_S_,
    't': PHASE_SHIFT_T,
    't_dg': PHASE_SHIFT_T_,
    'p': PHASE_SHIFT,
    'rx': ROTATION_X,
    'ry': ROTATION_Y,
    'rz': ROTATION_Z,
    'u1': ROTATION_U1,
    'u2': ROTATION_U2,
    'u3': ROTATION_U3,
    'cx': CONTROLLED_X,
    'cy': CONTROLLED_Y,
    'cz': CONTROLLED_Z,
    'cxr': CONTROLLED_XR,
    'cxr_dg': CONTROLLED_XR_,
    'ch': CONTROLLED_H,
    'cs': CONTROLLED_S,
    'cs_dg': CONTROLLED_S_,
    'ct': CONTROLLED_T,
    'ct_dg': CONTROLLED_T_,
    'cp': CONTROLLED_P,
    'crx': CONTROLLED_RX,
    'cry': CONTROLLED_RY,
    'crz': CONTROLLED_RZ,
    'cu1': CONTROLLED_U1,
    'cu2': CONTROLLED_U2,
    'cu3': CONTROLLED_U3,
    'sw': SWAP_QUBITS,
    'measure': MEASURE,
    'measure_x': MEASURE_X,
    'measure_y': MEASURE_Y,
    'measure_z': MEASURE_Z,
    'measure_bell': MEASURE_BELL,
    'reset': RESET,
}

GATE_STRING = {
    PAULI_X:'x',
    PAULI_Y:'y',
    PAULI_Z:'z',
    ROOT_PAULI_X:'xr',
    ROOT_PAULI_X_:'xr_dg',
    HADAMARD:'h',
    PHASE_SHIFT_S:'s',
    PHASE_SHIFT_S_:'s_dg',
    PHASE_SHIFT_T:'t',
    PHASE_SHIFT_T_:'t_dg',
    PHASE_SHIFT:'p',
    ROTATION_X:'rx',
    ROTATION_Y:'ry',
    ROTATION_Z:'rz',
    ROTATION_U1:'u1',
    ROTATION_U2:'u2',
    ROTATION_U3:'u3',
    CONTROLLED_X:'cx',
    CONTROLLED_Y:'cy',
    CONTROLLED_Z:'cz',
    CONTROLLED_XR:'cxr',
    CONTROLLED_XR_:'cxr_dg',
    CONTROLLED_H:'ch',
    CONTROLLED_S:'cs',
    CONTROLLED_S_:'cs_dg',
    CONTROLLED_T:'ct',
    CONTROLLED_T_:'ct_dg',
    CONTROLLED_P:'cp',
    CONTROLLED_RX:'crx',
    CONTROLLED_RY:'cry',
    CONTROLLED_RZ:'crz',
    CONTROLLED_U1:'cu1',
    CONTROLLED_U2:'cu2',
    CONTROLLED_U3:'cu3',
    SWAP_QUBITS:'sw',
    MEASURE:'measure',
    MEASURE_X:'measure_x',
    MEASURE_Y:'measure_y',
    MEASURE_Z:'measure_z',
    MEASURE_BELL:'meaxure_bell',
    RESET:'reset'
}

GATE_STRING_QASM = {
    # 1-qubit
    PAULI_X:'x',
    PAULI_Z:'z',
    HADAMARD:'h',
    PHASE_SHIFT_S:'s',
    PHASE_SHIFT_S_:'sdg',
    PHASE_SHIFT_T:'t',
    PHASE_SHIFT_T_:'tdg',
    RESET:'reset',
    # 1-qubit, 1-parameter
    ROTATION_X:'rx',
    ROTATION_Z:'rz',
    # 2-qubit
    CONTROLLED_X:'cx',
    CONTROLLED_Z:'cz',
    CONTROLLED_H:'ch',
    # 2-qubit, 1-parameter
    CONTROLLED_RZ:'crz',
    # measurement
    MEASURE:'measure',
}

# MatrixType

KRAUS = 1
POVM  = 2

# ApplyDir

LEFT  = 0
RIGHT = 1
BOTH  = 2

# ComplexAxis

REAL_PLUS  = 0
IMAG_PLUS  = 1
REAL_MINUS = 2
IMAG_MINUS = 3



# -*- coding: utf-8 -*-
from .QState import QState
from .Observable import Observable
from .DensOp import DensOp
from .Stabilizer import Stabilizer
from .Backend import Backend
from .QCirc import QCirc
from .CMem import CMem
from .PauliProduct import PauliProduct
from .Result import Result
from . import config
from . import error
from . import util

__all__ = ["QState","Observable","DensOp","Stabilizer","Backend","QCirc","CMem",
           "UnitaryOperator","PauliProduct","Result","config","error","util"]
